##########################################################################
# File:         IfParser.pm
#
# Description:  Perl module that implements the IfParser class
#               to retrive information from a XML interface definition file.
#
##########################################################################

package IfParser;

# To install the XML::LibXML module under Debian, apt-get libxml-libxml-perl
# Useful help on using this is in 'man XML::LibXML::Node'

use strict;
use XML::LibXML;

##########################################################################
# Class Constructor
##########################################################################
sub new
{
    # Standard Perl class constructor - see man perltoot(1)
    my $class = shift;

    # New takes one parameter, the name of the IF file to parse
    my $ifFile = shift;

    # Create the parser
    my $parser  = XML::LibXML->new() or die "Failed to create LibXML: $!";
    my $tree    = $parser->parse_file($ifFile) or die "Failed to parse $ifFile: $!";
    my $root    = $tree->getDocumentElement or die "Failed to get root object: $!";

    # Class instance data
    my $self = {
        PARSER    => $parser,
        XMLTREE   => $tree,
        ROOT      => $root,
    };

    bless($self, $class);
    return $self;
}

##########################################################################
# Public methods
##########################################################################

# Return an scalar containing the domain label attribute
sub getDomainLabel
{
    my $self = shift;
    my $domainLabel;

    my @domainNodes = $self->{ROOT}->findnodes("/Domain");

    if ($#domainNodes == 0)
    {
        my $label = $domainNodes[0]->findvalue('@label');
        if (defined $label and $label ne "")
        {
            $domainLabel = $label;
        }
    }

    return $domainLabel;
}

# Return an scalar containing the domain version attribute
sub getDomainVersion
{
    my $self = shift;
    my $domainVersion;

    my @domainNodes = $self->{ROOT}->findnodes("/Domain");

    if ($#domainNodes == 0)
    {
        my $version = $domainNodes[0]->findvalue('@version');
        if (defined $version and $version ne "")
        {
            $domainVersion = $version;
        }
    }

    return $domainVersion;
}

# Return a hash containing the domain limits attributes
sub getDomainLimits
{
    my $self = shift;
    my %domainLimits;

    my @domainNodes = $self->{ROOT}->findnodes("/Domain");

    if ($#domainNodes == 0)
    {
        $domainLimits{'lengthFieldSize'} = $domainNodes[0]->findvalue('@lengthFieldSize');
        $domainLimits{'tagFieldSize'} = $domainNodes[0]->findvalue('@tagFieldSize');
        $domainLimits{'maxMessageSize'} = $domainNodes[0]->findvalue('@maxMessageSize');
    }

    return %domainLimits;
}

# Return an scalar containing the packing used (defaults to maximal)
sub getDomainPacking
{
    my $self = shift;
    my $domainPacking;

    my @domainNodes = $self->{ROOT}->findnodes("/Domain");

    if ($#domainNodes == 0)
    {
        my $packing = $domainNodes[0]->findvalue('@packing');
        if (defined $packing and $packing ne "")
        {
            $domainPacking = $packing;
        }
        else
        {
            $domainPacking = "maximal";
        }
    }

    return $domainPacking;
}

# Return an array containing the names of all messages
sub getAllMessages
{
    my $self = shift;
    my @msgs;

    foreach my $msgNode ($self->{ROOT}->findnodes("/Domain/Message"))
    {
        my $msg = $msgNode->findvalue('@label');
        if (defined $msg and $msg ne "")
        {
            push @msgs, $msg;
        }
    }

    return @msgs;
}

# determines if the given syntax must be represented as an embedded IE
sub isEmbeddedIeType
{
    my $self = shift;
    my $syntax = shift;
    my $isEmbedded = 0;

    my @syntaxNode = $self->{ROOT}->findnodes("/Domain/SyntaxDefinition[\@label='".$syntax."']");
    if ($#syntaxNode == 0)
    {
        if ($syntaxNode[0]->find("ArrayType"))
        {
            my @element = $self->getArrayElementType($syntax);
            if (defined $element[1])
            {
                $isEmbedded = $self->isEmbeddedIeType($element[1]);
            }
            else
            {
                die("Array $syntax is missing element syntax");
            }
        }
        elsif ($syntaxNode[0]->find("SetType"))
        {
            my @element = $self->getSetElementType($syntax);
            if (defined $element[1])
            {
                $isEmbedded = $self->isEmbeddedIeType($element[1]);
            }
            else
            {
                die("Set $syntax is missing element syntax");
            }
        }
        elsif ($syntaxNode[0]->find("StructuredType"))
        {
            my @elements = $self->getAllStructInfoElements($syntax);
            foreach my $element (@elements)
            {
                if (defined $element and 
                    defined $$element{'syntax'} and 
                    $$element{'syntax'} ne "")
                {
                    if ($$element{'presence'} eq 'optional' or
                        $$element{'presence'} eq 'conditional')
                    {
                        $isEmbedded = 1;
                    } 
                    elsif ($self->isEmbeddedIeType($$element{'syntax'}))
                    {
                        $isEmbedded = 1;
                    }
                }
            }
        }
        else
        {
            # base data types do not need to be treated as embedded
            $isEmbedded = 0;
        }
    }
    else
    {
        die("More than one syntax definition of $syntax");
    }
    return $isEmbedded;
}

# Return an array containing the names of all syntax definitions
sub getAllSyntaxDefinitions
{
    my $self = shift;
    my @syntaxDefs;

    foreach my $syntaxNode ($self->{ROOT}->findnodes("/Domain/SyntaxDefinition"))
    {
        my $syntax = $syntaxNode->findvalue('@label');
        if (defined $syntax and $syntax ne "")
        {
            push @syntaxDefs, $syntax;
        }
    }

    return @syntaxDefs;
}

# Return the data type of the syntax definition
sub getSyntaxType
{
    my $self = shift;
    my $syntax = shift;
    my $syntaxType;

    my @syntaxNodes = $self->{ROOT}->findnodes("/Domain/SyntaxDefinition[\@label='".$syntax."']");

    if ($#syntaxNodes == 0)
    {
        my @syntaxTypes = $syntaxNodes[0]->findnodes('*');
        foreach my $syntax (@syntaxTypes)
        {
            if ($syntax->getName() ne "Usage" and $syntax->getName() ne "Reference")
            {
                $syntaxType = $syntax->getName();
            }
        }
    }
    return $syntaxType;
}

# Return an array containing the names of all syntax definitions for given type
sub getAllSyntaxByType
{
    my $self = shift;
    my $type = shift;
    my @syntaxDefs;

    foreach my $syntaxNode ($self->{ROOT}->findnodes("/Domain/SyntaxDefinition[$type]"))
    {
        my $syntax = $syntaxNode->findvalue('@label');
        if (defined $syntax and $syntax ne "")
        {
            push @syntaxDefs, $syntax;
        }
    }

    return @syntaxDefs;
}

# Return an array containing the names of all syntax definitions for structured types
sub getAllStructuredTypes
{
    my $self = shift;
    my @syntaxDefs = $self->getAllSyntaxByType("StructuredType");

    return @syntaxDefs;
}

# Return an array containing the syntaxes of all info elements defined
sub getAllInfoElements
{
    my $self = shift;
    my $type = shift;
    my $parentSyntax = shift;
    my %elements;

    my @infoElemNodes = $self->{ROOT}->findnodes("//InfoElement");

    foreach my $infoElem (@infoElemNodes)
    {
        my $syntax = $infoElem->findvalue('@syntax');
        if (defined $syntax and $syntax ne "")
        {
            # using a hash to avoid multiple instances of each syntax
            $elements{$syntax}="";
        }
    }

    return keys %elements;
}

# Return an array containing the info element label and syntax attributes, ordered by ordinal
sub getAllChildInfoElements
{
    my $self = shift;
    my $type = shift;
    my $parentSyntax = shift;
    my @elements;

    my $nodePath;

    if ($type eq "Message")
    {
        $nodePath = "/Domain/Message[\@label='".$parentSyntax."']/InfoElement";
    }
    elsif ($type eq "StructuredType")
    {
        $nodePath = "/Domain/SyntaxDefinition[\@label='".$parentSyntax."']/StructuredType/InfoElement";
    }
    my @infoElemNodes = $self->{ROOT}->findnodes($nodePath);

    foreach my $infoElem (@infoElemNodes)
    {
        my $ordinal = $infoElem->findvalue('@ordinal');
        if (defined $ordinal and $ordinal ne "")
        {
            my $syntax = $infoElem->findvalue('@syntax');
            if (defined $syntax and $syntax ne "")
            {
                my $label = $infoElem->findvalue('@label');
                if (!defined $label or $label eq "")
                {
                    $label = $syntax;
                }
                my $presence = $infoElem->findvalue('@presence');
                my $conditionalTest;
                my $conditionalTestLabel;
                if (!defined $presence or $presence eq "")
                {
                    $presence = "mandatory";
                }
                else
                {
                    if ($presence eq "conditional")
                    {
                        $conditionalTest = $infoElem->findvalue('ConditionalTest');
                        $conditionalTestLabel = $infoElem->findvalue('ConditionalTest/@label');
                    }
                }
                my $usage = $infoElem->findvalue('Usage');
                if (defined $usage and $usage ne "")
                {
                    # Strip any line feed chars from the string
                    $usage =~ s/\n/ /g;
                }
                my $reference = $infoElem->findvalue('Reference');
                if (defined $reference and $reference ne "")
                {
                    # Strip any line feed chars from the string
                    $reference =~ s/\n/ /g;
                }

                my %element;
                $element{'ordinal'} = $ordinal;
                $element{'syntax'} = $syntax;
                $element{'label'} = $label;
                $element{'presence'} = $presence;
                $element{'ConditionalTest'} = $conditionalTest;
                $element{'ConditionalTestLabel'} = $conditionalTestLabel;
                $element{'Usage'} = $usage;
                $element{'Reference'} = $reference;
                $elements[$ordinal] = \%element;
            }
        }
    }

    return @elements;
}

# Return an array containing the info element label and syntax attributes, ordered by ordinal
sub getChildInfoElement
{
    my $self = shift;
    my $type = shift;
    my $parentSyntax = shift;
    my $childSyntax = shift;
    my %element;

    my $nodePath;

    if ($type eq "Message")
    {
        $nodePath = "/Domain/Message[\@label='".$parentSyntax."']/InfoElement[\@syntax='".$childSyntax."']";
    }
    elsif ($type eq "StructuredType")
    {
        $nodePath = "/Domain/SyntaxDefinition[\@label='".$parentSyntax."']/StructuredType/InfoElement[\@syntax='".$childSyntax."']";
    }
    my @infoElemNodes = $self->{ROOT}->findnodes($nodePath);

    if ($#infoElemNodes == 0)
    {
        my $ordinal = $infoElemNodes[0]->findvalue('@ordinal');
        if (defined $ordinal and $ordinal ne "")
        {
            my $syntax = $infoElemNodes[0]->findvalue('@syntax');
            if (defined $syntax and $syntax ne "")
            {
                my $label = $infoElemNodes[0]->findvalue('@label');
                if (!defined $label or $label eq "")
                {
                    $label = $syntax;
                }
                my $presence = $infoElemNodes[0]->findvalue('@presence');
                my $conditionalTest;
                my $conditionalTestLabel;
                if (!defined $presence or $presence eq "")
                {
                    $presence = "mandatory";
                }
                else
                {
                    if ($presence eq "conditional")
                    {
                        $conditionalTest = $infoElemNodes[0]->findvalue('ConditionalTest');
                        $conditionalTestLabel = $infoElemNodes[0]->findvalue('ConditionalTest/@label');
                    }
                }
                my $usage = $infoElemNodes[0]->findvalue('Usage');
                if (defined $usage and $usage ne "")
                {
                    # Strip any line feed chars from the string
                    $usage =~ s/\n/ /g;
                }
                my $reference = $infoElemNodes[0]->findvalue('Reference');
                if (defined $reference and $reference ne "")
                {
                    # Strip any line feed chars from the string
                    $reference =~ s/\n/ /g;
                }

                $element{'ordinal'} = $ordinal;
                $element{'syntax'} = $syntax;
                $element{'label'} = $label;
                $element{'presence'} = $presence;
                $element{'ConditionalTest'} = $conditionalTest;
                $element{'ConditionalTestLabel'} = $conditionalTestLabel;
                $element{'Usage'} = $usage;
                $element{'Reference'} = $reference;
            }
        }
    }

    return %element;
}

sub getAllStructInfoElements
{
    my $self = shift;
    my $struct = shift;
    my @elements = $self->getAllChildInfoElements("StructuredType", $struct);

    return @elements;
}

sub getAllMessageInfoElements
{
    my $self = shift;
    my $message = shift;
    my @elements = $self->getAllChildInfoElements("Message", $message);

    return @elements;
}

# Return an array containing the names of all syntax definitions for UInt8 types
sub getAllUInt8Types
{
    my $self = shift;
    my @syntaxDefs = $self->getAllSyntaxByType("UInt8Type");

    return @syntaxDefs;
}

# Return an array containing the names of all syntax definitions for UInt16 types
sub getAllUInt16Types
{
    my $self = shift;
    my @syntaxDefs = $self->getAllSyntaxByType("UInt16Type");

    return @syntaxDefs;
}

# Return an array containing the names of all syntax definitions for UInt32 types
sub getAllUInt32Types
{
    my $self = shift;
    my @syntaxDefs = $self->getAllSyntaxByType("UInt32Type");

    return @syntaxDefs;
}

# Return an array containing the names of all syntax definitions for SInt8 types
sub getAllSInt8Types
{
    my $self = shift;
    my @syntaxDefs = $self->getAllSyntaxByType("SInt8Type");

    return @syntaxDefs;
}

# Return an array containing the names of all syntax definitions for SInt16 types
sub getAllSInt16Types
{
    my $self = shift;
    my @syntaxDefs = $self->getAllSyntaxByType("SInt16Type");

    return @syntaxDefs;
}

# Return an array containing the names of all syntax definitions for SInt32 types
sub getAllSInt32Types
{
    my $self = shift;
    my @syntaxDefs = $self->getAllSyntaxByType("SInt32Type");

    return @syntaxDefs;
}

# Return an array containing the element min and max attributes
sub getIntegerBounds
{
    my $self = shift;
    my $syntax = shift;
    my $intType = shift;

    return $self->getTypeBounds($syntax, $intType);
}

# Return an array containing the names of all syntax definitions for Boolean types
sub getAllBooleanTypes
{
    my $self = shift;
    my @syntaxDefs = $self->getAllSyntaxByType("BooleanType");

    return @syntaxDefs;
}

# Return an array containing the element label and syntax attributes
sub getElementType
{
    my $self = shift;
    my $container = shift;
    my $syntax = shift;
    my @element;

    my @arrayNodes = $self->{ROOT}->findnodes("/Domain/SyntaxDefinition[\@label='".$syntax."']/".$container."/Element");

    if ($#arrayNodes == 0)
    {
        my $syntax = $arrayNodes[0]->findvalue('@syntax');
        if (defined $syntax and $syntax ne "")
        {
            my $label = $arrayNodes[0]->findvalue('@label');
            if (defined $label and $label ne "")
            {
                @element = ($label, $syntax);
            }
            else
            {
                @element = ($syntax, $syntax);
            }
        }
    }

    return @element;
}

# Return an array containing the names of all syntax definitions for array types
sub getAllArrayTypes
{
    my $self = shift;
    my @syntaxDefs = $self->getAllSyntaxByType("ArrayType");

    return @syntaxDefs;
}

# Return an array containing the element label and syntax attributes
sub getArrayElementType
{
    my $self = shift;
    my $array = shift;

    return $self->getElementType("ArrayType", $array);
}

# Return an array containing the element minLength and maxLength attributes
sub getArrayBounds
{
    my $self = shift;
    my $syntax = shift;

    return $self->getTypeBounds($syntax, "ArrayType");
}

# Return an array containing the names of all syntax definitions for set types
sub getAllSetTypes
{
    my $self = shift;
    my @syntaxDefs = $self->getAllSyntaxByType("SetType");

    return @syntaxDefs;
}

# Return an array containing the element label and syntax attributes
sub getSetElementType
{
    my $self = shift;
    my $set = shift;

    return $self->getElementType("SetType", $set);
}

# Return an array containing the element minLength and maxLength attributes
sub getSetBounds
{
    my $self = shift;
    my $syntax = shift;

    return $self->getTypeBounds($syntax, "SetType");
}

# Return an array containing the names of all syntax definitions for version string types
sub getAllVersionStringTypes
{
    my $self = shift;
    my @syntaxDefs = $self->getAllSyntaxByType("VersionStringType");

    return @syntaxDefs;
}

# Return an array containing the names of all syntax definitions for string types
sub getAllStringTypes
{
    my $self = shift;
    my @syntaxDefs = $self->getAllSyntaxByType("StringType");

    return @syntaxDefs;
}

# Return an array containing the element minLength and maxLength attributes
sub getStringBounds
{
    my $self = shift;
    my $syntax = shift;

    return $self->getTypeBounds($syntax, "StringType");
}

# Return an array containing the names of all syntax definitions for octet array types
sub getAllOctetArrayTypes
{
    my $self = shift;
    my @syntaxDefs = $self->getAllSyntaxByType("OctetArrayType");

    return @syntaxDefs;
}

# Return an array containing the element minLength and maxLength attributes
sub getOctetArrayBounds
{
    my $self = shift;
    my $syntax = shift;

    return $self->getTypeBounds($syntax, "OctetArrayType");
}

# Return an array containing the element minLength and maxLength or min and max attributes
sub getTypeBounds
{
    my $self = shift;
    my $syntax = shift;
    my $type = shift;
    my @bounds;

    my @syntaxNodes = $self->{ROOT}->findnodes("/Domain/SyntaxDefinition[\@label='".$syntax."']/$type");

    if ($#syntaxNodes == 0)
    {
        my $max = $syntaxNodes[0]->findvalue('@maxLength');
        if (defined $max and $max ne "")
        {
            my $min = $syntaxNodes[0]->findvalue('@minLength');
            if (!defined $min or $min eq "")
            {
                $min = 0;
            }

            @bounds = ($min, $max);
        }
        else
        {
            my $max = $syntaxNodes[0]->findvalue('@max');
            if (!defined $max or $max eq "")
            {
                if ($type eq "SInt8Type")
                {
                    $max = "127";
                }
                elsif ($type eq "SInt16Type")
                {
                    $max = "32767";
                }
                elsif ($type eq "SInt32Type")
                {
                    $max = "2147483647";
                }
                elsif ($type eq "UInt8Type")
                {
                    $max = "255";
                }
                elsif ($type eq "UInt16Type")
                {
                    $max = "65535";
                }
                elsif ($type eq "UInt32Type")
                {
                    $max = "-1";
                }
            }
            my $min = $syntaxNodes[0]->findvalue('@min');
            if (!defined $min or $min eq "")
            {
                if ($type eq "SInt8Type")
                {
                    $min = "-128";
                }
                elsif ($type eq "SInt16Type")
                {
                    $min = "-32768";
                }
                elsif ($type eq "SInt32Type")
                {
                    $min = "-2147483648";
                }
                elsif ($type eq "UInt8Type")
                {
                    $min = "0";
                }
                elsif ($type eq "UInt16Type")
                {
                    $min = "0";
                }
                elsif ($type eq "UInt32Type")
                {
                    $min = "0";
                }
            }
            @bounds = ($min, $max);
        }
    }

    return @bounds;
}

# Return an array containing the names of all syntax definitions for enum types
sub getAllEnumTypes
{
    my $self = shift;
    my @syntaxDefs = $self->getAllSyntaxByType("EnumerationType");

    return @syntaxDefs;
}

# Return an array containing the details of all the enum values
sub getAllEnumTypeValues
{
    my $self = shift;
    my $syntax = shift;
    my @values;

    my @valueNodes = $self->{ROOT}->findnodes("/Domain/SyntaxDefinition[\@label='".$syntax."']/EnumerationType/Value");

    foreach my $value (@valueNodes)
    {
        my $ordinal = $value->findvalue('@ordinal');
        if (defined $ordinal and $ordinal ne "")
        {
            my $enum = $value->findvalue('@enum');
            if (defined $enum and $enum ne "")
            {
                my $display = $value->findvalue('@displayName');
                if (defined $display and $display ne "")
                {
                    my %element;
                    $element{'ordinal'} = $ordinal;
                    $element{'enum'} = $enum;
                    $element{'displayName'} = $display;
                    $values[$ordinal] = \%element;
                }
            }
        }
    }

    return @values;
}

# Return an array containing the element displayName, usage and reference attributes
sub getSyntaxDescriptions
{
    my $self = shift;
    my $syntax = shift;
    my @descriptions;

    my @syntaxNodes = $self->{ROOT}->findnodes("/Domain/SyntaxDefinition[\@label='".$syntax."']");

    if ($#syntaxNodes == 0)
    {
        my $display = $syntaxNodes[0]->findvalue('@displayName');
        if (defined $display and $display ne "")
        {
            push(@descriptions, $display);
            
            my $usage = $syntaxNodes[0]->findvalue('Usage');
            if (defined $usage and $usage ne "")
            {
                # Strip any line feed chars from the string
                $usage =~ s/\n/ /g;

                push(@descriptions, $usage);
            }

            my $reference = $syntaxNodes[0]->findvalue('Reference');
            if (defined $reference and $reference ne "")
            {
                # Strip any line feed chars from the string
                $reference =~ s/\n/ /g;

                push(@descriptions, $reference);
            }
        }
    }

    return @descriptions;
}

# Return a hash containing the syntax attributes
sub getSyntaxAttributes
{
    my $self = shift;
    my $tag = shift;
    my %syntaxAttrs;

    my @syntaxNodes = $self->{ROOT}->findnodes("/Domain/SyntaxDefinition[\@tag='".$tag."']");

    if ($#syntaxNodes == 0)
    {
        $syntaxAttrs{'tag'} = $tag;

        my $display = $syntaxNodes[0]->findvalue('@displayName');
        if (defined $display and $display ne "")
        {
            $syntaxAttrs{'displayName'} = $display;
        }            

        my $label = $syntaxNodes[0]->findvalue('@label');
        if (defined $label and $label ne "")
        {
            $syntaxAttrs{'label'} = $label;
        }
    }

    return %syntaxAttrs;
}

# Return an array containing the message displayName, usage and reference attributes
sub getMessageDescriptions
{
    my $self = shift;
    my $msg = shift;
    my @descriptions;

    my @msgNodes = $self->{ROOT}->findnodes("/Domain/Message[\@label='".$msg."']");

    if ($#msgNodes == 0)
    {
        my $display = $msgNodes[0]->findvalue('@displayName');
        if (defined $display and $display ne "")
        {
            push(@descriptions, $display);
            
            my $usage = $msgNodes[0]->findvalue('Usage');
            if (defined $usage and $usage ne "")
            {
                push(@descriptions, $usage);
            }

            my $reference = $msgNodes[0]->findvalue('Reference');
            if (defined $reference and $reference ne "")
            {
                push(@descriptions, $reference);
            }
        }
    }

    return @descriptions;
}

# Return a hash containing the message attributes
sub getMessageAttributes
{
    my $self = shift;
    my $msgId = shift;
    my %msgAttrs;

    my @msgNodes = $self->{ROOT}->findnodes("/Domain/Message[\@messageId='".$msgId."']");

    if ($#msgNodes == 0)
    {
        $msgAttrs{'messageId'} = $msgId;

        my $display = $msgNodes[0]->findvalue('@displayName');
        if (defined $display and $display ne "")
        {
            $msgAttrs{'displayName'} = $display;
        }            

        my $label = $msgNodes[0]->findvalue('@label');
        if (defined $label and $label ne "")
        {
            $msgAttrs{'label'} = $label;
        }
    }

    return %msgAttrs;
}

# Return a scalar containing the messageId
sub getMessageId
{
    my $self = shift;
    my $msg = shift;
    my $msgId = 0;

    my @msgNodes = $self->{ROOT}->findnodes("/Domain/Message[\@label='".$msg."']");

    if ($#msgNodes == 0)
    {
        $msgId = $msgNodes[0]->findvalue('@messageId');
    }

    return $msgId;
}

# Return a scalar containing the tag of a syntax definition
sub getSyntaxDefinitionTag
{
    my $self = shift;
    my $syntax = shift;
    my $tag = 0;

    my @syntaxNodes = $self->{ROOT}->findnodes("/Domain/SyntaxDefinition[\@label='".$syntax."']");

    if ($#syntaxNodes == 0)
    {
        $tag = $syntaxNodes[0]->findvalue('@tag');
    }

    return $tag;
}

# Given a string, return a multiline string as a C/C++ comment.
# Lines are limited to a maximum line length of $maxLength characters.
# Takes 4 params:
#     - prefix string to use at the start of each line (e.g. "    /* ")
#     - postfix string to use at the end of each line (e.g. " */")
#     - max line length
#     - the string to turn into a comment
sub makeComment
{
    my $prefix    = shift;
    my $postfix   = shift;
    my $maxLength = shift;
    my $text      = shift;
    my $comment   = "";
    my $lineLimit = $maxLength - length($postfix);

    if ($text)
    {
        # Strip any line feed chars from the string
        $text =~ s/\n/ /g;

        # Strip any leading whitespace from the string
        $text =~ s/^\s*//;

        # Convert any tabs to spaces
        $text =~ s/\t/ /g;

        # Start line with indent spaces and comment chars
        my $nextLine = $prefix;

        foreach my $word (split / /, $text)
        {
            # Can the next word fit on the line?
            if (length($nextLine) + length($word) < $lineLimit)
            {
                # Don't forget to put the space back in between words!
                $nextLine .= $word." ";
            }
            else
            {
                while (length($nextLine) < $lineLimit) { $nextLine .= " "; }
                $comment .= $nextLine.$postfix."\n";
                # Stick it on the next line
                $nextLine = $prefix.$word." ";
            }
        }

        while (length($nextLine) < $lineLimit) { $nextLine .= " "; }
        $comment .= $nextLine.$postfix."\n";
    }

    return $comment;
}

##########################################################################
# Must be last in class file - return true!
##########################################################################
1;

