#!/usr/bin/perl

##########################################################################
# File:         xmlToC.pl
#
# Description:  Script to parse the XML interface definition and 
#               generate C code
#
#  Copyright (C) 2006 ip.access Ltd
#  All Rights Reserved
#
##########################################################################

use strict;

# Add the current directory to the modules include path
use lib ".";

use devParser;
use IfUtils;

# tool version
my $xmlToWsdVersion = '$Revision: 1.3 $';
# strip $Revision from text
$xmlToWsdVersion =~ s/\$//g;
$xmlToWsdVersion =~ s/Revision: //;
# major version number (initially added to distinguish move to new archive)
$xmlToWsdVersion = "2." . $xmlToWsdVersion;

# Limit for recursive protocols
my $maxDepthLimit = 100;

# Usage
if ($#ARGV != 0)
{
    print "Usage: xmlToWireSharkDissector.pl [XML File]\n\n";
    print "Version: $xmlToWsdVersion\n\n";
    print "Note: The following features are NOT supported:\n";
    print "    Uniqueness testing of Sets\n";
    print "    UniString\n";
    print "    String pattern testing checking\n";
    print "    Reduced code base with InterfaceUser\n";
    print "    Minimal domain encoding\n";
    exit(1);
}

# The XML IF file
my $ifFile = $ARGV[0];

##########################################################################
# Main program
##########################################################################

my $parser = devParser->new($ifFile);

my $domain = $parser->getApplicationLabel();
my $templateCFileName = mkName("packet-oyster3g-${domain}.c");
my $templateHFileName = mkName("packet-oyster3g-${domain}.h");
my $template;

my %domainLimits = $parser->getApplicationLimits();

my $lengthSizeType = "UInt32";
if ($domainLimits{'lengthFieldSize'} == 1)
{
    $lengthSizeType = "UInt8";
}
elsif ($domainLimits{'lengthFieldSize'} == 2)
{
    $lengthSizeType = "UInt16";
}
elsif ($domainLimits{'lengthFieldSize'} == 4)
{
    $lengthSizeType = "UInt32";
}

my $tagSizeType = "UInt32";
if ($domainLimits{'tagFieldSize'} == 1)
{
    $tagSizeType = "UInt8";
}
elsif ($domainLimits{'tagFieldSize'} == 2)
{
    $tagSizeType = "UInt16";
}
elsif ($domainLimits{'tagFieldSize'} == 4)
{
    $tagSizeType = "UInt32";
}

# ref to message Id Ie type, must be first element in all messages
my $refDeviceIdType;

# pre load syntax info
my @allSyntaxes  = $parser->getAllSyntaxDefinitions();
my %syntaxIndex = makeSyntaxIndexHash(\@allSyntaxes);

my @allDevices  = $parser->getAllDevices();

generateTemplateHeader($templateHFileName);
generateApplicationId();
generateDeviceIds();
generateSyntaxIds();
generateTemplateFooter($templateHFileName);

generateTemplateHeader($templateCFileName);
generateCTemplateHeader();
generateDomainIeVars();
generateSyntaxHfs();
generateEnumTypes();
generateDeviceIeArrays();
generateDeviceStructs();
generateSyntaxIeArrays();
generateSyntaxStructs();
generateEttVars();
generateMsgPtrArray();
generateSyntaxPtrArray();
generateDomainDefnStruct();
generateRegisterInfo();
generateCTemplateFooter();
generateTemplateFooter($templateCFileName);

exit 0;

##########################################################################
# Subroutines
##########################################################################

sub makeSyntaxIndexHash
{
    my $syntaxes = shift;
    my %syntaxIndexHash;

    for (my $index = 0; $index < scalar(@$syntaxes); $index++)
    {
        $syntaxIndexHash{$$syntaxes[$index]} = $index;
    }
    return %syntaxIndexHash;
}

sub calculateNestingDepth
{
    my $maxDepth = 1;

    foreach my $syntax (@allSyntaxes)
    {
        my $depth = 1;
        $depth = getSyntaxDepth($syntax, $depth);
        if ($depth > $maxDepth)
        {
            $maxDepth = $depth;
        }
    }
    return $maxDepth;
}

sub getSyntaxDepth
{
    my $syntax = shift;
    my $maxDepth = shift;
    $maxDepth++;

    if ($maxDepth > $maxDepthLimit)
    {
        return $maxDepthLimit;
    }

    if ($parser->getSyntaxType($syntax) eq "StructuredType")
    {
        my $newMaxDepth = $maxDepth;
        my @infoElems = $parser->getAllStructInfoElements($syntax);
        for (my $ordinal=0; $ordinal <= $#infoElems; $ordinal++)
        {
            my $depth = $maxDepth;
            my $element = $infoElems[$ordinal];
            if (defined $element and defined $$element{'syntax'})
            {
                $depth = getSyntaxDepth($$element{'syntax'}, $depth);
                if ($depth > $newMaxDepth)
                {
                    $newMaxDepth = $depth;
                }
            }
        }
        $maxDepth = $newMaxDepth;
    }
    elsif ($parser->getSyntaxType($syntax) eq "SetType")
    {
        my @element = $parser->getSetElementType($syntax);
        my ($elemName, $elemSyntax) = @element;
        if (defined $elemSyntax)
        {
            $maxDepth = getSyntaxDepth($elemSyntax, $maxDepth);
        }
    }
    elsif ($parser->getSyntaxType($syntax) eq "ArrayType")
    {
        my @element = $parser->getArrayElementType($syntax);
        my ($elemName, $elemSyntax) = @element;
        if (defined $elemSyntax)
        {
            $maxDepth = getSyntaxDepth($elemSyntax, $maxDepth);
        }
    }
    
    return $maxDepth;
}

#
# Code generation
#

sub generateEttVars
{
    print "/* the number of the following element handles and the corresponding array is the maximum \"depth\" of nesting\n";
    print " * of set/array and structured syntax elements + 1 (for the protocol message itself)\n";
    print " * It may be set to a \"safe\" large value, but ... */\n";
    my $depth = calculateNestingDepth();
    for (my $level = 0; $level < $depth; $level++)
    {
        print "static gint ".mkName("ett",$domain,$level)." = -1;\n";
    }
    print "static gint *ett[] = \n";
    print "{\n";
    for (my $level = 0; $level < $depth; $level++)
    {
        print "    &".mkName("ett",$domain,$level).",\n";
    }
    print "};\n";
}

sub generateMsgPtrArray
{
    print "/* The following section is auto-generated from the XML Device definitions.\n";
    print " * It consistes of an array for the message definitions above.\n";
    print " * The name of should be of the form \"<domain label>_message\".\n";
    print " */\n";

    print "static const oyster3g_device_t* ".mkName($domain,"message")."[] = \n";
    print "{\n";
    foreach my $msg (@allDevices)
    {
        print "    &".mkName($msg,"msg").",\n";
    }
    print "    NULL\n";
    print "};\n";
    print "\n";
}

sub generateSyntaxPtrArray
{
    print "/* The following section is auto-generated from the XML Syntax definitions.\n";
    print " * It consistes of an array for the syntax definitions above.\n";
    print " * The name of should be of the form \"<domain label>_syntax\".\n";
    print " */\n";

    print "static const oyster3g_device_syntax_t* ".mkName($domain,"syntax")."[] = \n";
    print "{\n";
    foreach my $syntax (@allSyntaxes)
    {
        print "    &".mkName($syntax,"syntax").",\n";
    }
    print "    NULL\n";
    print "};\n";
    print "\n";
    print "/* The following section is auto-generated from the XML Syntax definitions.\n";
    print " * It consistes of an array for customisation functions for the syntax definitions above.\n";
    print " * The name of should be of the form \"<domain label>_custom_syntax\".\n";
    print " */\n";

    print "static custom_dissector_t ".mkName($domain,"custom_syntax")."[] = \n";
    print "{\n";
    foreach my $syntax (@allSyntaxes)
    {
        print "    NULL,\n";
    }
    print "    NULL\n";
    print "};\n";
    print "\n";
}

sub generateDomainDefnStruct
{
    print "/* The following section is auto-generated from the XML Domain definition.\n";
    print " * It consistes of a structure which defines the domain, including arrays\n";
    print " * for the message and syntax definitions above.\n";
    print " * The name of domain structure should be of the form \"<domain label>_domain\".\n";
    print " */\n";

    print "static const struct oyster3g_device_domain_t ".mkName($domain,"domain")." = {\n";
    print "    \"$domain\",             /* label of the Domain */\n";
    print "    \"Oyster3G $domain Domain\",      /* Display Name of the Domain */\n";
    print "    ".$domainLimits{'lengthFieldSize'}.",                  /* size of length fields */\n";
    print "    ".$domainLimits{'tagFieldSize'}.",                  /* size of tag fields */\n";
    print "    TRUE,               /* TRUE if maximal encoding used on the wire */\n";
    print "    ".mkName($domain,"message").",       /* array of pointers to message definitions */\n";
    print "    ".mkName($domain,"syntax").",        /* array of pointers to syntax definitions */\n";
    print "                        /* below this point are the wireshark handles for\n";
    print "                        * this domain, initialised to the above statics.\n";
    print "                        */\n";
    print "    ".mkName($domain,"custom_syntax").",        /* array of pointers to custom syntax dissectors */\n";
    print "    &".mkName("proto",$domain).",       /* proto handle */\n";
    print "    array_length(ett), /* the size of the ett array */\n";
    print "    ett,               /* ett array handle */\n";
    print "    &hf_ie_tag,        /* the hf value for the (hidden) IE tag field */\n";
    print "    &hf_ie_len,        /* the hf value for the (hidden) IE length field */\n";
    print "};\n";
    print "\n";
}

sub generateRegisterInfo
{
    print "/* The following section is auto-generated from the XML Syntax definitions.\n";
    print " * It consists of an array of wireshark \"hf\" structures, one per syntax which is a base type.\n";
    print " * The handle variables are defined above.\n";
    print " * The \"hf\" integer is a unique handle used by wireshark to display and name data of this type.\n";
    print " * The name of the hf variable should be of the form \"hf_<syntax label>\"\n";
    print " * The representations FT* are mapped from the Syntax type:\n";
    print " *  FT_BOOLEAN, FT_UINT8, FT_UINT16, FT_UINT32, FT_INT8, FT_INT16 and FT_INT32 from the obvious base types\n";
    print " *  FT_UINT16 is also used for our Enums\n";
    print " *  FT_STRING from both String and Version String base types\n";
    print " *  FT_BYTES from OCTET_ARRAY\n";
    print " *  FT_NONE from Sets, Arrays and Structs\n";
    print " * There is currently no useful mapping for Unicode strings - use FT_BYTES\n";
    print " * The base of various integer should be BASE_DEC_HEX (as we have no clues in the XML)\n";
    print " * The base of booleans should be BASE_NONE as we only have full octet booleans\n";
    print " * The base of all other types should be BASE_NONE\n";
    print " * The strings field is only used for Enums and should be \"VALS(<syntax label>_vals)\", and should be NULL for the rest\n";
    print " */\n";
    print "\n";
    print "static hf_register_info hf[] = \n";
    print "{\n";
    print "    { &hf_ie_tag,\n";
    print "        { \"Attribute Tag\", \"oyster3g.".mkName($domain).".ie_tag\", FT_".uc($tagSizeType).", BASE_DEC_HEX, NULL,\n";
    print "        0x0, \"Attribute ID Field\", HFILL }\n";
    print "    },\n";
    print "    { &hf_ie_len,\n";
    print "        { \"Attribute Len\", \"oyster3g.".mkName($domain).".ie_len\", FT_".uc($lengthSizeType).", BASE_DEC_HEX, NULL,\n";
    print "        0x0, \"Attribute Length Field\", HFILL }\n";
    print "    },\n";
    print "\n";
    foreach my $syntax (@allSyntaxes)
    {
        my @descriptions = $parser->getSyntaxDescriptions($syntax);
        my ($displayName, $usage, $reference) = @descriptions;

        print "    {\n";
        print "        &".mkName("hf",$syntax).",\n";
        print "        {\n";
        print "            \"$displayName\",\n";
        print "            \"oyster3g.".mkName($domain).".".mkName($syntax)."\",\n";

        if ($parser->getSyntaxType($syntax) eq "StructuredType")
        {
            print "            FT_NONE,\n";
            print "            BASE_NONE,\n";
            print "            NULL,\n";
        }
        elsif ($parser->getSyntaxType($syntax) eq "SetType")
        {
            print "            FT_NONE,\n";
            print "            BASE_NONE,\n";
            print "            NULL,\n";
        }
        elsif ($parser->getSyntaxType($syntax) eq "ArrayType")
        {
            print "            FT_NONE,\n";
            print "            BASE_NONE,\n";
            print "            NULL,\n";
        }
        elsif ($parser->getSyntaxType($syntax) eq "BooleanType")
        {
            print "            FT_BOOLEAN,\n";
            print "            BASE_NONE,\n";
            print "            NULL,\n";
        }
        elsif ($parser->getSyntaxType($syntax) eq "EnumerationType")
        {
            print "            FT_UINT16,\n";
            print "            BASE_DEC_HEX,\n";
            print "            VALS(".mkName($syntax,"vals")."),\n";
        }
        elsif ($parser->getSyntaxType($syntax) eq "SInt8Type")
        {
            print "            FT_INT8,\n";
            print "            BASE_DEC,\n";
            print "            NULL,\n";
        }
        elsif ($parser->getSyntaxType($syntax) eq "SInt16Type")
        {
            print "            FT_INT16,\n";
            print "            BASE_DEC,\n";
            print "            NULL,\n";
        }
        elsif ($parser->getSyntaxType($syntax) eq "SInt32Type")
        {
            print "            FT_INT32,\n";
            print "            BASE_DEC,\n";
            print "            NULL,\n";
        }
        elsif ($parser->getSyntaxType($syntax) eq "UInt8Type")
        {
            print "            FT_UINT8,\n";
            print "            BASE_DEC_HEX,\n";
            print "            NULL,\n";
        }
        elsif ($parser->getSyntaxType($syntax) eq "UInt16Type")
        {
            print "            FT_UINT16,\n";
            print "            BASE_DEC_HEX,\n";
            print "            NULL,\n";
        }
        elsif ($parser->getSyntaxType($syntax) eq "UInt32Type")
        {
            print "            FT_UINT32,\n";
            print "            BASE_DEC_HEX,\n";
            print "            NULL,\n";
        }
        elsif ($parser->getSyntaxType($syntax) eq "StringType" or 
               $parser->getSyntaxType($syntax) eq "VersionStringType")
        {
            print "            FT_STRING,\n";
            print "            BASE_NONE,\n";
            print "            NULL,\n";
        }
        elsif ($parser->getSyntaxType($syntax) eq "UniStringType")
        {
            print "            FT_BYTES,\n";
            print "            BASE_NONE,\n";
            print "            NULL,\n";
        }
        elsif ($parser->getSyntaxType($syntax) eq "OctetArrayType")
        {
            print "            FT_BYTES,\n";
            print "            BASE_NONE,\n";
            print "            NULL,\n";
        }
        else
        {
            print STDERR "Unknown syntax type $syntax==".$parser->getSyntaxType($syntax)."\n";
        }
        print "            0x0,\n";
        print "            \"$displayName\",\n";
        print "            HFILL\n";
        print "        }\n";
        print "    },\n";
    }
    print "};\n";
    print "\n";
}

sub generateEnumTypes
{
    my @allEnums  = $parser->getAllEnumTypes();

    print "/* The following section is auto-generated from the XML Syntax definitions with base type of enumeration\n";
    print " * It consists a set of value string array definitions.\n";
    print " * The name of the value string should be of the form \"<Syntax label>_vals\".\n";
    print " * Each array entry consists of the enumeration constant followed by the Display Name (if present)\n";
    print " * or label (if no Display Name) associated with that string value.\n";
    print " * Each value_string array ends in a NULL entry {0, NULL}.\n";
    print " */\n";

    foreach my $syntax (@allEnums)
    {
        my @values = $parser->getAllEnumTypeValues($syntax);

        print "static const value_string ".mkName($syntax,"vals")."[] = \n";
        print "{\n";
        foreach my $value (@values)
        {
            if (!defined $$value{'displayName'} or $$value{'displayName'} eq "")
            {
                $$value{'displayName'} = $$value{'enum'};
            }
            if (defined $$value{'ordinal'})
            {
	            print "    { ".$$value{'ordinal'}.", \"".$$value{'displayName'}."\" },\n";
            }
        }
        print "    { 0, NULL } /* Null entry to terminate list */\n";
        print "};\n";
        print "\n";
    }
}

sub generateSyntaxHfs
{
    print "/* The following section is auto-generated from the XML syntax definitions.\n";
    print " * It consistes of a set of statics, one per syntax, for the wireshark field\n";
    print " * registrations.\n";
    print " * The name of the static should be of the form \"hf_<SYNTAX_LABEL>\".\n";
    print " * The value is initialised to -1.\n";
    print " */\n";
    print "\n";

    foreach my $syntax (@allSyntaxes)
    {
        print "static int ".mkName("hf",$syntax)." = -1;\n";
    }
    print "\n";
}

sub generateDomainIeVars
{
    print "/* The following entries are for the protocol and the basic IE tag and length\n";
    print " * field registrations used by wireshark.\n";
    print " * The protocol field has the name \"proto_<domain label>\".\n";
    print " * The values are initialised to -1.\n";
    print " */\n";
    print "\n";
    print "static int ".mkName("proto",$domain)." = -1;\n";
    print "\n";
    print "static int hf_ie_tag = -1;\n";
    print "static int hf_ie_len = -1;\n";
    print "\n";
}

sub generateCommonDefines
{
    print "/*********************************************************\n";
    print " * ${domain} common definitions\n";
    print " *********************************************************/\n";

    my $line = "#define ".enumName("MAX_MESSAGE_SIZE")." ";
    while (length($line) < 50) { $line .= " "; }        
    $line .= $domainLimits{'maxDeviceSize'}."\n";
    print $line;
    
    $line = "#define ".enumName("NUM_TAG_FIELD_OCTETS")." ";
    while (length($line) < 50) { $line .= " "; }        
    $line .= $domainLimits{'tagFieldSize'}."\n";
    print $line;
    
    $line = "#define ".enumName("NUM_LENGTH_FIELD_OCTETS")." ";
    while (length($line) < 50) { $line .= " "; }        
    $line .= $domainLimits{'lengthFieldSize'}."\n";
    print $line;
    
    print "\n";
}

sub generateApplicationId
{
    print "/* The following section is auto-generated from the XML message definitions.\n";
    print " * It is a \#define for the application Id of this whole application.\n";
    print " * The name of the \#define should be of the form \"<APPLICATION_LABEL>_ID\".\n";
    print " * The value is the Application Id tag value for the Application.\n";
    print " */\n";

    my $line = "#define ".defName($domain,"ID")." ";
    while (length($line) < 50) { $line .= " "; }        
    $line .= $parser->getApplicationId()."\n";
    print $line;
    print "\n";
}

sub generateDeviceIds
{
    print "/* The following section is auto-generated from the XML message definitions.\n";
    print " * It consistes of a set of \#defines, one per messsage.\n";
    print " * The name of the \#define should be of the form \"<MESSAGE_LABEL>_ID\".\n";
    print " * The value is the message Id tag value for the message.\n";
    print " */\n";

    foreach my $msg (@allDevices)
    {
        my $line = "#define ".defName($msg,"ID")." ";
        while (length($line) < 50) { $line .= " "; }        
        $line .= $parser->getDeviceType($msg)."\n";
        print $line;
    }
    print "\n";
}

sub generateSyntaxIds
{
    print "/* The following section is auto-generated from the XML syntax definitions.\n";
    print " * It consistes of a set of \#defines, one per syntax definition.\n";
    print " * The name of the \#define should be of the form \"<SYNTAX_LABEL>_SYNTAX\".\n";
    print " * The value is the tag value for the syntax.\n";
    print " */\n";

    foreach my $syntax (@allSyntaxes)
    {
        my $line = "#define ".defName($syntax,"SYNTAX")." ";
        while (length($line) < 50) { $line .= " "; }        
        $line .= $parser->getSyntaxDefinitionTag($syntax)."\n";
        print $line;
    }
    print "\n";
}

sub generateDeviceIeArrays
{
    print "/* The following section is auto-generated from the XML Device IE definitions.\n";
    print " * It consistes of a set of message IE arrays.\n";
    print " * The name of message IE array should be of the form \"<message label>_ie\".\n";
    print " * The array of info element definitions ends in the Null info element entry\n";
    print " * {NULL, -1, -1, FALSE}.\n";
    print " * NOTE: the Syntax element of the IE structure is the ARRAY INDEX in the Syntax array below\n";
    print " * which corresponds to the Syntax for this IE.\n";
    print " */\n";

    foreach my $msg (@allDevices)
    {
        print "static const struct oyster3g_device_ie_t ".mkName($msg,"ie")."[] = \n";
        print "{\n";
        my @infoElems = $parser->getAllDeviceAttributes($msg);
        for (my $ordinal=0; $ordinal <= $#infoElems; $ordinal++)
        {
            my $element = $infoElems[$ordinal];
            if (defined $element)
            {
                print "    { \"".$$element{'label'}."\", ".$$element{'id'}.", ".$syntaxIndex{$$element{'syntax'}};
                if ($$element{'presence'} eq "mandatory")
                {
                    print ", TRUE },\n";
                }
                else
                {
                    print ", FALSE },\n";
                }
            }
        }
        print "    { NULL, -1, -1, FALSE } /* Null entry to terminate list */\n";
        print "};\n";
    }
    print "\n";
}

sub generateDeviceStructs
{
    print "/* The following section is auto-generated from the XML Device definitions.\n";
    print " * It consistes of a set of message structures, one per messsage.\n";
    print " * The name of message structure should be of the form \"<message label>_msg\".\n";
    print " * Each message includes a pointer to the array of info element definitions above.\n";
    print " */\n";

    foreach my $msg (@allDevices)
    {
        my @descriptions = $parser->getDeviceDescriptions($msg);
        my ($displayName, $usage, $reference) = @descriptions;

        print "static const struct oyster3g_device_t ".mkName($msg,"msg")." = \n";
        print "{\n";
        print "    \"$msg\",\n";
        print "    \"$displayName\",\n";
        print "    ".defName($msg,"ID").",\n";
        print "    ".mkName($msg,"ie")."\n";
        print "};\n";
    }
    print "\n";
}

sub generateSyntaxIeArrays
{
    print "/* The following section is auto-generated from the XML Syntax definitions.\n";
    print " * It consistes of a set of syntax IE arrays, one per non-base syntax.\n";
    print " * The name of syntax structure should be of the form \"<syntax label>_ie\".\n";
    print " * The array takes one for the following forms:\n";
    print " *  For Sets and Arrays the array has a single entry for the element definition of the Set/Array\n";
    print " *  For Structs the array is a set of entries, one per info element in the struucture definition,\n";
    print " *  ending in a Null info element entry {NULL, -1, -1, FALSE}.\n";
    print " */\n";

    foreach my $syntax (@allSyntaxes)
    {
        if ($parser->getSyntaxType($syntax) eq "StructuredType")
        {
            print "static const struct oyster3g_device_ie_t ".mkName($syntax,"ie")."[] = \n";
            print "{\n";
            my @infoElems = $parser->getAllStructInfoElements($syntax);
            for (my $ordinal=0; $ordinal <= $#infoElems; $ordinal++)
            {
                my $element = $infoElems[$ordinal];
                if (defined $element)
                {
                    print "    { \"".$$element{'label'}."\", -1, ".$syntaxIndex{$$element{'syntax'}};
                    if ($$element{'presence'} eq "mandatory")
                    {
                        print ", TRUE },\n";
                    }
                    else
                    {
                        print ", FALSE },\n";
                    }
                }
            }
            print "    { NULL, -1, -1, FALSE } /* Null entry to terminate list */\n";
            print "};\n";
        }
        elsif ($parser->getSyntaxType($syntax) eq "SetType")
        {
            print "static const struct oyster3g_device_ie_t ".mkName($syntax,"ie")."[] = \n";
            print "{\n";
            my @element = $parser->getSetElementType($syntax);
            my ($elemName, $elemSyntax) = @element;
            if (defined $elemSyntax)
            {
                print "    { \"".$elemName."\", -1, ".$syntaxIndex{$elemSyntax}.", TRUE },\n";
            }
            print "    { NULL, -1, -1, FALSE } /* Null entry to terminate list */\n";
            print "};\n";
        }
        elsif ($parser->getSyntaxType($syntax) eq "ArrayType")
        {
            print "static const struct oyster3g_device_ie_t ".mkName($syntax,"ie")."[] = \n";
            print "{\n";
            my @element = $parser->getArrayElementType($syntax);
            my ($elemName, $elemSyntax) = @element;
            if (defined $elemSyntax)
            {
                print "    { \"".$elemName."\", -1, ".$syntaxIndex{$elemSyntax}.", TRUE },\n";
            }
            print "    { NULL, -1, -1, FALSE } /* Null entry to terminate list */\n";
            print "};\n";
        }
    }
    print "\n";
}

sub generateSyntaxStructs
{
    print "/* The following section is auto-generated from the XML Syntax definitions.\n";
    print " * It consistes of a set of syntax structures, one per syntax.\n";
    print " * The name of syntax structure should be of the form \"<syntax label>_syntax\".\n";
    print " * Each syntax includes a pointer to the hf element definitions above.\n";
    print " * Compound syntax definitions include reference to the ie which defines it.\n";
    print " */\n";

    foreach my $syntax (@allSyntaxes)
    {
        my @descriptions = $parser->getSyntaxDescriptions($syntax);
        my ($displayName, $usage, $reference) = @descriptions;

        print "static const struct oyster3g_device_syntax_t ".mkName($syntax,"syntax")." = \n";
        print "{\n";
        print "    \"$syntax\",\n";
        print "    ".defName($syntax,"SYNTAX").",\n";
        print "    \"$displayName\",\n";

        if ($parser->getSyntaxType($syntax) eq "StructuredType")
        {
            print "    OYSTER3G_SYNTAX_STRUCT,\n";
            print "    ".mkName($syntax,"ie").",\n";
        }
        elsif ($parser->getSyntaxType($syntax) eq "SetType")
        {
            print "    OYSTER3G_SYNTAX_SET,\n";
            print "    ".mkName($syntax,"ie").",\n";
        }
        elsif ($parser->getSyntaxType($syntax) eq "ArrayType")
        {
            print "    OYSTER3G_SYNTAX_ARRAY,\n";
            print "    ".mkName($syntax,"ie").",\n";
        }
        elsif ($parser->getSyntaxType($syntax) eq "BooleanType")
        {
            print "    OYSTER3G_SYNTAX_BOOLEAN,\n";
            print "    NULL,\n";
        }
        elsif ($parser->getSyntaxType($syntax) eq "EnumerationType")
        {
            print "    OYSTER3G_SYNTAX_ENUM,\n";
            print "    NULL,\n";
        }
        elsif ($parser->getSyntaxType($syntax) eq "SInt8Type")
        {
            print "    OYSTER3G_SYNTAX_SINT8,\n";
            print "    NULL,\n";
        }
        elsif ($parser->getSyntaxType($syntax) eq "SInt16Type")
        {
            print "    OYSTER3G_SYNTAX_SINT16,\n";
            print "    NULL,\n";
        }
        elsif ($parser->getSyntaxType($syntax) eq "SInt32Type")
        {
            print "    OYSTER3G_SYNTAX_SINT32,\n";
            print "    NULL,\n";
        }
        elsif ($parser->getSyntaxType($syntax) eq "UInt8Type")
        {
            print "    OYSTER3G_SYNTAX_UINT8,\n";
            print "    NULL,\n";
        }
        elsif ($parser->getSyntaxType($syntax) eq "UInt16Type")
        {
            print "    OYSTER3G_SYNTAX_UINT16,\n";
            print "    NULL,\n";
        }
        elsif ($parser->getSyntaxType($syntax) eq "UInt32Type")
        {
            print "    OYSTER3G_SYNTAX_UINT32,\n";
            print "    NULL,\n";
        }
        elsif ($parser->getSyntaxType($syntax) eq "StringType" or 
               $parser->getSyntaxType($syntax) eq "VersionStringType")
        {
            print "    OYSTER3G_SYNTAX_STRING,\n";
            print "    NULL,\n";
        }
        elsif ($parser->getSyntaxType($syntax) eq "UniStringType")
        {
            print "    OYSTER3G_SYNTAX_UNI_STRING,\n";
            print "    NULL,\n";
        }
        elsif ($parser->getSyntaxType($syntax) eq "OctetArrayType")
        {
            print "    OYSTER3G_SYNTAX_OCTET_ARRAY,\n";
            print "    NULL,\n";
        }
        else
        {
            print STDERR "Unknown syntax type $syntax==".$parser->getSyntaxType($syntax)."\n";
        }
        print "    &".mkName("hf",$syntax)."\n";
        print "};\n";
    }
    print "\n";
}

# takes a list of hash refs to child syntax definitions and returns another 
# list with the list of children of unique syntax, order is maintained for 
# unique values and the first instance of any duplicated child.
# sparse arrays will be compressed
sub uniqueChildList
{
    my @inputList = @_;
    my @outputList;
    my %values;
    my $outputIndex = 0;
    for (my $index=0; $index < scalar(@inputList); $index++)
    {
        if (defined $inputList[$index])
        {
            if (!defined $values{$inputList[$index]->{'syntax'}})
            {
                $values{$inputList[$index]->{'syntax'}} = [$outputIndex, $inputList[$index]];
                $outputIndex++;
            }
        }
    }
    foreach my $value (keys %values)
    {
        $outputList[$values{$value}->[0]] = $values{$value}->[1];
    }
    return @outputList;
}

# takes a list of scalars and returns another list with the
# list of values being unique, order is maintained for 
# unique values and the first instance of any duplicated values.
# sparse arrays will be compressed
sub uniqueScalarList
{
    my @inputList = @_;
    my @outputList;
    my %values;
    my $outputIndex = 0;
    for (my $index=0; $index < scalar(@inputList); $index++)
    {
        if (defined $inputList[$index])
        {
            if (!defined $values{$inputList[$index]})
            {
                $values{$inputList[$index]} = $outputIndex;
                $outputIndex++;
            }
        }
    }
    foreach my $value (keys %values)
    {
        $outputList[$values{$value}] = $value;
    }
    return @outputList;
}

##########################################################################
# Generated code templates
##########################################################################

sub generateTemplateHeader
{
    my $templateFileName = shift;
    
    open($template, ">$templateFileName") or die("Cannot open $templateFileName: $!");

    select($template);

    print "/* $templateFileName\n";
    print " * Routines for oyster3g $domain packet disassembly\n";
    print " *\n";
    print " * \$"; # split over two lines to avoid MKS replacing the tag
    print "Id \$ \n";
    print " *\n";
    print " * Copyright (c) 2007 by ip.access Ltd <".'support@ipaccess.com'.">\n";
    print " *\n";
    print " * Wireshark - Network traffic analyzer\n";
    print " * By Gerald Combs <".'gerald@wireshark.org'.">\n";
    print " * Copyright 1999 Gerald Combs\n";
    print " *\n";
    print " * This program is free software; you can redistribute it and/or\n";
    print " * modify it under the terms of the GNU General Public License\n";
    print " * as published by the Free Software Foundation; either version 2\n";
    print " * of the License, or (at your option) any later version.\n";
    print " *\n";
    print " * This program is distributed in the hope that it will be useful,\n";
    print " * but WITHOUT ANY WARRANTY; without even the implied warranty of\n";
    print " * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n";
    print " * GNU General Public License for more details.\n";
    print " *\n";
    print " * You should have received a copy of the GNU General Public License\n";
    print " * along with this program; if not, write to the Free Software\n";
    print " * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.\n";
    print " */\n";
    print "\n";
    print "/*\n";
    print " *  WireShark Dissector autogenerated from the XML definition.\n";
    print " *\n";
    print " *  DO NOT EDIT THIS FILE\n";
    print " *\n";
    print " *  Autogen version: $xmlToWsdVersion\n";
    print " */\n";
    print "\n";
}

sub generateCTemplateHeader
{
    print "/* Include files */\n";
    print "\n";
    print "#ifdef HAVE_CONFIG_H\n";
    print "#include \"config.h\"\n";
    print "#endif\n";
    print "\n";
    print "#include <stdlib.h>\n";
    print "#include <ctype.h>\n";
    print "#include <epan/packet.h>\n";
    print "#include <epan/prefs.h>\n";
    print "#include <epan/strutil.h>\n";
    print "\n";
    print "/* The following include defines the structures used for the static data\n";
    print " * definitions which follow\n";
    print " */\n";
    print "\n";
    print "#include \"packet-device-oyster3g.h\"\n";
    print "#include \"$templateHFileName\"\n";
    print "\n";
}

sub generateCTemplateFooter
{
    print "\n";
    print "/* the main dissector */\n";
    print "static void ".mkName("dissect_oyster3g",$domain)."(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree)\n";
    print "{\n";
    print "    /* simply call the generic domain dissector with the domain definition and state struct */\n";
    print "    dissect_oyster3g_device_domain(tvb, pinfo, main_tree, &".mkName($domain,"domain").");\n";
    print "    \n";
    print "    return;\n";
    print "}\n";
    print "\n";
    print "/* the registration routine is called at startup */\n";
    print "void ".mkName("proto_register_oyster3g",$domain)."(void) \n";
    print "{\n";
    print "    ".mkName("proto",$domain)." = proto_register_protocol(\"Oyster3G ${domain} Domain\",\"Oyster3G ${domain}\",\"".mkName("o3g",${domain})."\");\n";
    print "    register_dissector(\"".mkName("o3g",${domain})."\", ".mkName("dissect_oyster3g",$domain).", ".mkName("proto",$domain).");\n";
    print "\n";
    print "    proto_register_field_array(".mkName("proto",$domain).",hf,array_length(hf));\n";
    print "\n";
    print "    proto_register_subtree_array(ett,array_length(ett));\n";
    print "}\n";
    print "\n";
    print "/* The registration hand-off routine is called after all dissectors have been initialised individually */\n";
    print "void ".mkName("proto_reg_handoff_oyster3g",$domain)."(void) \n";
    print "{\n";
    print "    int i;\n";
    print "    static int ".mkName($domain,"initialized")." = FALSE;\n";
    print "    static dissector_handle_t ".mkName($domain,"handle").";\n";
    print "\n";
    print "    if(!".mkName($domain,"initialized").") \n";
    print "    {\n";
    print "        ".mkName($domain,"handle")." = create_dissector_handle(".mkName("dissect_oyster3g",$domain).", ".mkName("proto",$domain).");\n";
    print "        ".mkName($domain,"initialized")." = TRUE;\n";
    print "    }\n";
    print "    /* The following \#define is manually created if needed in packet-oyster3g.h\n";
    print "     * It allows this domain protocol to be chained to the SOIP dissector\n";
    print "     */\n";
    print "    else \n";
    print "    {\n";
    print "        for (i=0; ".mkName($domain,"message")."[i]; i++)\n";
    print "           dissector_delete(\"o3g_device_interface.device\", (".defName($domain,"ID")."<<16)|".mkName($domain,"message")."[i]->tag, ".mkName($domain,"handle").");\n";
    print "    }\n";
    print "\n";
    print "    for (i=0; ".mkName($domain,"message")."[i]; i++)\n";
    print "       dissector_add(\"o3g_device_interface.device\", (".defName($domain,"ID")."<<16)|".mkName($domain,"message")."[i]->tag, ".mkName($domain,"handle").");\n";
    print "\n";
    print "/* The registration of custom dissectors for this domain */\n";
    print "#ifdef ".defName("OYSTER3G",$domain,"CUSTOM_DISSECTORS")."\n";
    print "    ".mkName("oyster3g_register_custom_dissectors",$domain)."(&".mkName($domain,"domain").");\n";
    print "#endif\n";
    print "}\n";
    print "\n";
}

sub generateTemplateFooter
{
    my $templateFileName = shift;

    print "\n";

    close($template);
    select(STDOUT);

    print "Generated Source: $templateFileName\n";
}

##########################################################################
# Name mangling utils
##########################################################################

sub mkName
{
    my @parts = @_;
    my $name = shift @parts;

    foreach my $part (@parts)
    {
        $name .= "_" . $part;
    }
    $name =~ s/([a-z])([A-Z])/\1_\2/g;
    $name = lc $name;
    
    return $name;
}

sub defName
{
    my @parts = @_;
    my $defName = lc shift @parts;

    foreach my $part (@parts)
    {
        $defName .= ucfirst $part;
    }
    $defName =~ s/([a-z0-9])([A-Z])/\1_\2/g;
    $defName = uc($defName);
    
    return $defName;
}

sub doAndDie
{
    my $text = shift;
    select(STDOUT);

    die("ERROR: $text");
}


