##########################################################################
# File:         devParser.pm
#
# Description:  Perl module that implements the devParser class
#               to retrive information from a XML device type definition file.
#
##########################################################################

package devParser;

# To install the XML::LibXML module under Debian, apt-get libxml-libxml-perl
# Useful help on using this is in 'man XML::LibXML::Node'

use strict;
use XML::LibXML;

##########################################################################
# Class Constructor
##########################################################################
sub new
{
    # Standard Perl class constructor - see man perltoot(1)
    my $class = shift;

    # New takes one parameter, the name of the IF file to parse
    my $devFile = shift;

    # Create the parser
    my $parser  = XML::LibXML->new() or die "Failed to create LibXML: $!";
    my $tree    = $parser->parse_file($devFile) or die "Failed to parse $devFile: $!";
    my $root    = $tree->getDocumentElement or die "Failed to get root object: $!";       
         
    # Class instance data
    my $self = {
        PARSER    => $parser,
        XMLTREE   => $tree,
        ROOT      => $root,
    };

    bless($self, $class);
    return $self;
}

##########################################################################
# Public methods
##########################################################################

# Return an scalar containing the application label attribute
sub getApplicationLabel
{
    my $self = shift;
    my $applicationLabel;   
  
    my @applicationNodes = $self->{ROOT}->findnodes("/Application");

    if ($#applicationNodes == 0)
    {
        my $label = $applicationNodes[0]->findvalue('@label');
        if (defined $label and $label ne "")
        {
            $applicationLabel = $label;
        }
    }

    return $applicationLabel;
}

# Return an scalar containing the application version attribute
sub getApplicationVersion
{
    my $self = shift;
    my $applicationVersion;

    my @applicationNodes = $self->{ROOT}->findnodes("/Application");

    if ($#applicationNodes == 0)
    {
        my $version = $applicationNodes[0]->findvalue('@version');
        if (defined $version and $version ne "")
        {
            $applicationVersion = $version;
        }
    }

    return $applicationVersion;
}

# Return an scalar containing the application version attribute
sub getApplicationId
{
    my $self = shift;
    my $applicationId;

    my @applicationNodes = $self->{ROOT}->findnodes("/Application");

    if ($#applicationNodes == 0)
    {
        my $appId = $applicationNodes[0]->findvalue('@appId');
        if (defined $appId and $appId ne "")
        {
            $applicationId = $appId;
        }
    }

    return $applicationId;
}

# Return a hash containing the application limits attributes
sub getApplicationLimits
{
    my $self = shift;
    my %applicationLimits;

    my @applicationNodes = $self->{ROOT}->findnodes("/Application");

    if ($#applicationNodes == 0)
    {
        $applicationLimits{'lengthFieldSize'} = $applicationNodes[0]->findvalue('@lengthFieldSize');
        $applicationLimits{'tagFieldSize'} = $applicationNodes[0]->findvalue('@tagFieldSize');
	$applicationLimits{'maxDeviceSize'} = $applicationNodes[0]->findvalue('@maxDeviceSize');
    }

    return %applicationLimits;
}

# Return an array containing the names of all devices
sub getAllDevices
{
    my $self = shift;
    my @devs;

    foreach my $devNode ($self->{ROOT}->findnodes("/Application/Device"))
    {
        my $dev = $devNode->findvalue('@label');
        if (defined $dev and $dev ne "")
        {
            push @devs, $dev;
        }
    }

    return @devs;
}

# Return an array containing the names of all syntax definitions
sub getAllSyntaxDefinitions
{
    my $self = shift;
    my @syntaxDefs;

    foreach my $syntaxNode ($self->{ROOT}->findnodes("/Application/SyntaxDefinition"))
    {
        my $syntax = $syntaxNode->findvalue('@label');
        if (defined $syntax and $syntax ne "")
        {
            push @syntaxDefs, $syntax;
        }
    }

    return @syntaxDefs;
}

# Return the data type of the syntax definition
sub getSyntaxType
{
    my $self = shift;
    my $syntax = shift;
    my $syntaxType;

    my @syntaxNodes = $self->{ROOT}->findnodes("/Application/SyntaxDefinition[\@label='".$syntax."']");

    if ($#syntaxNodes == 0)
    {
        my @syntaxTypes = $syntaxNodes[0]->findnodes('*');
        foreach my $syntax (@syntaxTypes)
        {
            if ($syntax->getName() ne "Usage" and $syntax->getName() ne "Reference")
            {
                $syntaxType = $syntax->getName();
            }
        }
    }
    return $syntaxType;
}

# Return an array containing the names of all syntax definitions for given type
sub getAllSyntaxByType
{
    my $self = shift;
    my $type = shift;
    my @syntaxDefs;

    foreach my $syntaxNode ($self->{ROOT}->findnodes("/Application/SyntaxDefinition[$type]"))
    {
        my $syntax = $syntaxNode->findvalue('@label');
        if (defined $syntax and $syntax ne "")
        {
            push @syntaxDefs, $syntax;
        }
    }

    return @syntaxDefs;
}

# Return an array containing the labels of all info elements defined
sub getAllAttributeLabels
{
    my $self = shift;
    my %elements;

    my @attribNodes = $self->{ROOT}->findnodes("//DevAttribute");

    foreach my $attribs (@attribNodes)
    {
        my $label = $attribs->findvalue('@label');
        if (defined $label and $label ne "")
        {
            # using a hash to avoid multiple instances of each label
            $elements{$label}="";
        }
    }

    return keys %elements;
}

# Return an attribute syntax based on the label 
sub getAttributeSyntax
{
    my $self = shift;
    my $label = shift;
    my $syntax;

    my @attribNodes = $self->{ROOT}->findnodes("//DevAttribute[\@label='".$label."']");

    $syntax = $attribNodes[0]->findvalue('@syntax');

    return $syntax;
}

# Return an array containing the info element label and syntax attributes, ordered by id
sub getAllChildDeviceAttribs
{
    my $self = shift;
    my $parentSyntax = shift;
    my @elements;

    my $nodePath = "/Application/Device[\@label='".$parentSyntax."']/DevAttribute";
    
    my @devAttribNodes = $self->{ROOT}->findnodes($nodePath);

    my $index = 0;
    foreach my $devAttrib (@devAttribNodes)
    {
        my $id = $devAttrib->findvalue('@id');
        if (defined $id and $id ne "")
        {
            my $label = $devAttrib->findvalue('@label');
            if (defined $label and $label ne "")
            {
                my $syntax = $devAttrib->findvalue('@syntax');
                my $presence = $devAttrib->findvalue('@presence');
                my $conditionalTest;
                my $conditionalTestLabel;
                if (!defined $presence or $presence eq "")
                {
                    $presence = "mandatory";
                }
                else
                {
                    if ($presence eq "conditional")
                    {
                        $conditionalTest = $devAttrib->findvalue('ConditionalTest');
                        $conditionalTestLabel = $devAttrib->findvalue('ConditionalTest/@label');
                    }
                }
                my $usage = $devAttrib->findvalue('Usage');
                if (defined $usage and $usage ne "")
                {
                    # Strip any line feed chars from the string
                    $usage =~ s/\n/ /g;
                }
                my $reference = $devAttrib->findvalue('Reference');
                if (defined $reference and $reference ne "")
                {
                    # Strip any line feed chars from the string
                    $reference =~ s/\n/ /g;
                }

                my $modifiable = $devAttrib->findvalue('@modifiable');

                my %element;
                $element{'id'} = $id;
                $element{'syntax'} = $syntax;
                $element{'label'} = $label;
                $element{'presence'} = $presence;
                $element{'ConditionalTest'} = $conditionalTest;
                $element{'ConditionalTestLabel'} = $conditionalTestLabel;
                $element{'Usage'} = $usage;
                $element{'Reference'} = $reference;
                $element{'modifiable'} = $modifiable;

                $elements[$index] = \%element;
                ++$index;
            }
        }
    }

    return @elements;
}

sub getAllDeviceAttributes
{
    my $self = shift;
    my $device = shift;
    my @elements = $self->getAllChildDeviceAttribs($device);

    return @elements;
}

# Return an array containing the names of all syntax definitions for UInt8 types
sub getAllUInt8Types
{
    my $self = shift;
    my @syntaxDefs = $self->getAllSyntaxByType("UInt8Type");

    return @syntaxDefs;
}

# Return an array containing the names of all syntax definitions for UInt16 types
sub getAllUInt16Types
{
    my $self = shift;
    my @syntaxDefs = $self->getAllSyntaxByType("UInt16Type");

    return @syntaxDefs;
}

# Return an array containing the names of all syntax definitions for UInt32 types
sub getAllUInt32Types
{
    my $self = shift;
    my @syntaxDefs = $self->getAllSyntaxByType("UInt32Type");

    return @syntaxDefs;
}

# Return an array containing the names of all syntax definitions for SInt8 types
sub getAllSInt8Types
{
    my $self = shift;
    my @syntaxDefs = $self->getAllSyntaxByType("SInt8Type");

    return @syntaxDefs;
}

# Return an array containing the names of all syntax definitions for SInt16 types
sub getAllSInt16Types
{
    my $self = shift;
    my @syntaxDefs = $self->getAllSyntaxByType("SInt16Type");

    return @syntaxDefs;
}

# Return an array containing the names of all syntax definitions for SInt32 types
sub getAllSInt32Types
{
    my $self = shift;
    my @syntaxDefs = $self->getAllSyntaxByType("SInt32Type");

    return @syntaxDefs;
}

# Return an array containing the element min and max attributes
sub getIntegerBounds
{
    my $self = shift;
    my $syntax = shift;
    my $intType = shift;

    return $self->getTypeBounds($syntax, $intType);
}

# Return an array containing the names of all syntax definitions for Boolean types
sub getAllBooleanTypes
{
    my $self = shift;
    my @syntaxDefs = $self->getAllSyntaxByType("BooleanType");

    return @syntaxDefs;
}

# Return an array containing the element label and syntax attributes
sub getElementType
{
    my $self = shift;
    my $container = shift;
    my $syntax = shift;
    my @element;

    my @arrayNodes = $self->{ROOT}->findnodes("/Application/SyntaxDefinition[\@label='".$syntax."']/".$container."/Element");

    if ($#arrayNodes == 0)
    {
        my $syntax = $arrayNodes[0]->findvalue('@syntax');
        if (defined $syntax and $syntax ne "")
        {
            my $label = $arrayNodes[0]->findvalue('@label');
            if (defined $label and $label ne "")
            {
                @element = ($label, $syntax);
            }
            else
            {
                @element = ($syntax, $syntax);
            }
        }
    }

    return @element;
}

# Return an array containing the names of all syntax definitions for array types
sub getAllArrayTypes
{
    my $self = shift;
    my @syntaxDefs = $self->getAllSyntaxByType("ArrayType");

    return @syntaxDefs;
}

# Return an array containing the element label and syntax attributes
sub getArrayElementType
{
    my $self = shift;
    my $array = shift;

    return $self->getElementType("ArrayType", $array);
}

# Return an array containing the element minLength and maxLength attributes
sub getArrayBounds
{
    my $self = shift;
    my $syntax = shift;

    return $self->getTypeBounds($syntax, "ArrayType");
}

# Return an array containing the names of all syntax definitions for version string types
sub getAllVersionStringTypes
{
    my $self = shift;
    my @syntaxDefs = $self->getAllSyntaxByType("VersionStringType");

    return @syntaxDefs;
}

# Return an array containing the names of all syntax definitions for string types
sub getAllStringTypes
{
    my $self = shift;
    my @syntaxDefs = $self->getAllSyntaxByType("StringType");

    return @syntaxDefs;
}

# Return an array containing the element minLength and maxLength attributes
sub getStringBounds
{
    my $self = shift;
    my $syntax = shift;

    return $self->getTypeBounds($syntax, "StringType");
}

# Return an array containing the names of all syntax definitions for octet array types
sub getAllOctetArrayTypes
{
    my $self = shift;
    my @syntaxDefs = $self->getAllSyntaxByType("OctetArrayType");

    return @syntaxDefs;
}

# Return an array containing the element minLength and maxLength attributes
sub getOctetArrayBounds
{
    my $self = shift;
    my $syntax = shift;

    return $self->getTypeBounds($syntax, "OctetArrayType");
}

# Return an array containing the element minLength and maxLength attributes
sub getTypeBounds
{
    my $self = shift;
    my $syntax = shift;
    my $type = shift;
    my @bounds;

    my @syntaxNodes = $self->{ROOT}->findnodes("/Application/SyntaxDefinition[\@label='".$syntax."']/$type");

    if ($#syntaxNodes == 0)
    {
        my $max = $syntaxNodes[0]->findvalue('@maxLength');
        if (defined $max and $max ne "")
        {
            my $min = $syntaxNodes[0]->findvalue('@minLength');
            if (!defined $min or $min eq "")
            {
                $min = 0;
            }

            @bounds = ($min, $max);
        }
        else
        {
            my $max = $syntaxNodes[0]->findvalue('@max');
            if (defined $max and $max ne "")
            {
                my $min = $syntaxNodes[0]->findvalue('@min');
                if (!defined $min or $min eq "")
                {
                    $min = 0;
                }

                @bounds = ($min, $max);
            }
        }
    }

    return @bounds;
}


# Return an array containing the names of all syntax definitions for enum types
sub getAllEnumTypes
{
    my $self = shift;
    my @syntaxDefs = $self->getAllSyntaxByType("EnumerationType");

    return @syntaxDefs;
}

# Return an array containing the details of all the enum values
sub getAllEnumTypeValues
{
    my $self = shift;
    my $syntax = shift;
    my @values;

    my @valueNodes = $self->{ROOT}->findnodes("/Application/SyntaxDefinition[\@label='".$syntax."']/EnumerationType/Value");

    foreach my $value (@valueNodes)
    {
        my $ordinal = $value->findvalue('@ordinal');
        if (defined $ordinal and $ordinal ne "")
        {
            my $enum = $value->findvalue('@enum');
            if (defined $enum and $enum ne "")
            {
                my $display = $value->findvalue('@displayName');
                if (defined $display and $display ne "")
                {
                    my %element;
                    $element{'ordinal'} = $ordinal;
                    $element{'enum'} = $enum;
                    $element{'displayName'} = $display;
                    $values[$ordinal] = \%element;
                }
            }
        }
    }

    return @values;
}

# Return an array containing the element displayName, usage and reference attributes
sub getSyntaxDescriptions
{
    my $self = shift;
    my $syntax = shift;
    my @descriptions;

    my @syntaxNodes = $self->{ROOT}->findnodes("/Application/SyntaxDefinition[\@label='".$syntax."']");

    if ($#syntaxNodes == 0)
    {
        my $display = $syntaxNodes[0]->findvalue('@displayName');
        if (defined $display and $display ne "")
        {
            push(@descriptions, $display);
            
            my $usage = $syntaxNodes[0]->findvalue('Usage');
            if (defined $usage and $usage ne "")
            {
                # Strip any line feed chars from the string
                $usage =~ s/\n/ /g;

                push(@descriptions, $usage);
            }

            my $reference = $syntaxNodes[0]->findvalue('Reference');
            if (defined $reference and $reference ne "")
            {
                # Strip any line feed chars from the string
                $reference =~ s/\n/ /g;

                push(@descriptions, $reference);
            }
        }
    }

    return @descriptions;
}

# Return a hash containing the syntax attributes
sub getSyntaxAttributes
{
    my $self = shift;
    my $tag = shift;
    my %syntaxAttrs;

    my @syntaxNodes = $self->{ROOT}->findnodes("/Application/SyntaxDefinition[\@tag='".$tag."']");

    if ($#syntaxNodes == 0)
    {
        $syntaxAttrs{'tag'} = $tag;

        my $display = $syntaxNodes[0]->findvalue('@displayName');
        if (defined $display and $display ne "")
        {
            $syntaxAttrs{'displayName'} = $display;
        }            

        my $label = $syntaxNodes[0]->findvalue('@label');
        if (defined $label and $label ne "")
        {
            $syntaxAttrs{'label'} = $label;
        }
    }

    return %syntaxAttrs;
}

sub getDeviceDescriptions
{
    my $self = shift;
    my $dev = shift;
    my @descriptions;

    my @devNodes = $self->{ROOT}->findnodes("/Application/Device[\@label='".$dev."']");

    if ($#devNodes == 0)
    {
        my $display = $devNodes[0]->findvalue('@displayName');
        if (defined $display and $display ne "")
        {
            push(@descriptions, $display);
            
            my $usage = $devNodes[0]->findvalue('Usage');
            if (defined $usage and $usage ne "")
            {
                push(@descriptions, $usage);
            }

            my $reference = $devNodes[0]->findvalue('Reference');
            if (defined $reference and $reference ne "")
            {
                push(@descriptions, $reference);
            }
        }
    }

    return @descriptions;
}

sub getDeviceType
{
    my $self = shift;
    my $dev = shift;
    my $devId = 0;

    my @devNodes = $self->{ROOT}->findnodes("/Application/Device[\@label='".$dev."']");

    if ($#devNodes == 0)
    {
        $devId = $devNodes[0]->findvalue('@deviceType');
    }

    return $devId;
}
# Return a scalar containing the tag of a syntax definition
sub getSyntaxDefinitionTag
{
    my $self = shift;
    my $syntax = shift;
    my $tag = 0;

    my @syntaxNodes = $self->{ROOT}->findnodes("/Application/SyntaxDefinition[\@label='".$syntax."']");

    if ($#syntaxNodes == 0)
    {
        $tag = $syntaxNodes[0]->findvalue('@tag');
    }

    return $tag;
}

# Return a string containing the display name of a syntax definition
sub getSyntaxDefinitionDisplay
{
    my $self = shift;
    my $syntax = shift;
    my $display = 0;

    my @syntaxNodes = $self->{ROOT}->findnodes("/Application/SyntaxDefinition[\@label='".$syntax."']");

    if ($#syntaxNodes == 0)
    {
        $display = $syntaxNodes[0]->findvalue('@displayName');
    }

    return $display;
}

# Given a string, return a multiline string as a C/C++ comment.
# Lines are limited to a maximum line length of $maxLength characters.
# Takes 4 params:
#     - prefix string to use at the start of each line (e.g. "    /* ")
#     - postfix string to use at the end of each line (e.g. " */")
#     - max line length
#     - the string to turn into a comment
sub makeComment
{
    my $prefix    = shift;
    my $postfix   = shift;
    my $maxLength = shift;
    my $text      = shift;
    my $comment   = "";
    my $lineLimit = $maxLength - length($postfix);

    if ($text)
    {
        # Strip any line feed chars from the string
        $text =~ s/\n/ /g;

        # Strip any leading whitespace from the string
        $text =~ s/^\s*//;

        # Convert any tabs to spaces
        $text =~ s/\t/ /g;

        # Start line with indent spaces and comment chars
        my $nextLine = $prefix;

        foreach my $word (split / /, $text)
        {
            # Can the next word fit on the line?
            if (length($nextLine) + length($word) < $lineLimit)
            {
                # Don't forget to put the space back in between words!
                $nextLine .= $word." ";
            }
            else
            {
                while (length($nextLine) < $lineLimit) { $nextLine .= " "; }
                $comment .= $nextLine.$postfix."\n";
                # Stick it on the next line
                $nextLine = $prefix.$word." ";
            }
        }

        while (length($nextLine) < $lineLimit) { $nextLine .= " "; }
        $comment .= $nextLine.$postfix."\n";
    }

    return $comment;
}

##########################################################################
# Must be last in class file - return true!
##########################################################################
1;

