/* packet-oyster3g-custom-ursl.c
 * Custom Routines for oyster3g URSL packet disassembly
 *
 * $Id $
 *
 * Copyright (c) 2007 by ip.access Ltd <support@ipaccess.com>
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1999 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/* Include files */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <ctype.h>
#include <epan/packet.h>
#include <epan/prefs.h>
#include <epan/strutil.h>
#include <epan/dissectors/packet-rtp.h>
#include <epan/dissectors/packet-rtcp.h>

/* The following include defines the structures used for the static data
 * definitions which follow
 */

#include "packet-oyster3g.h"
#include "packet-oyster3g-ursl.h"

static struct {
unsigned char ac_rtp_ip_address[4];
unsigned char curr_addr_index_no;
int ac_rtp_port;
guint packet_num;
address  ac_rtp_ip_addr;
address  ac_rtcp_ip_addr;
GHashTable *rtp_dyn_payload ;
} g_rtp_stream_info;

static unsigned char current_message_id;
static dissector_handle_t nas_handle=NULL;
static dissector_handle_t srnc_to_trnc_handle=NULL;
static dissector_handle_t trnc_to_srnc_handle=NULL;
static dissector_handle_t trnc_to_inter_rat_handle=NULL;


static int dissect_ursl_acipaddr(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree);
static int dissect_ursl_acudpport(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree);
static int dissect_ursl_rtpserverinfo(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree);

static int dissect_ursl_nasmsg(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree);
static int dissect_rrc_container(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree);
static int dissect_message_id(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree);

static custom_syntax_dissector_t dissector[] =
{
	{NASMSG_SYNTAX, &dissect_ursl_nasmsg},
	{RTPSERVERINFO_SYNTAX, &dissect_ursl_rtpserverinfo},
	{IPADDRESSBYTE_SYNTAX, &dissect_ursl_acipaddr},
	{UDPPORT_SYNTAX, &dissect_ursl_acudpport},
	{RRCCONTAINER_SYNTAX, &dissect_rrc_container},
	{MESSAGEID_SYNTAX, &dissect_message_id},
	{0,NULL}
};

void oyster3g_register_custom_dissectors_ursl(const oyster3g_domain_t *domain)
{
    return;
}

int dissect_message_id(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree){
    current_message_id = tvb_get_guint8(tvb,0);
    return 0;
}

void oyster3g_register_custom_dissectors_handoff_ursl(const oyster3g_domain_t *domain)
{
    gint* key,*key1,*key2,*key3;
    dissector_handle_t ip_handle;

    oyster3g_register_custom_dissectors(domain, dissector);

    nas_handle = find_dissector("gsm_a_dtap");

    ip_handle = find_dissector("ip");

    srnc_to_trnc_handle = find_dissector("rrc.s_to_trnc_cont");
    trnc_to_srnc_handle = find_dissector("rrc.t_to_srnc_cont");
    trnc_to_inter_rat_handle = find_dissector("rrc.t_to_interrat_cont");

    dissector_add_string("rtp_dyn_payload_type","RTP_IP", ip_handle);

    g_rtp_stream_info.curr_addr_index_no = 0;

    g_rtp_stream_info.rtp_dyn_payload = g_hash_table_new( g_int_hash, g_int_equal);
    key = g_malloc(sizeof(gint));
    *key = 97; /* 97 is used for CS Video **/
    g_hash_table_insert(g_rtp_stream_info.rtp_dyn_payload,key,"CLEARMODE");

    key1 = g_malloc(sizeof(gint));
    *key1 = 96;
    g_hash_table_insert(g_rtp_stream_info.rtp_dyn_payload,key1,"VND.3GPP.IUFP");

    key2 = g_malloc(sizeof(gint));
    *key2 = 101;
    g_hash_table_insert(g_rtp_stream_info.rtp_dyn_payload,key2,"VND.3GPP.IUFP");

    key3 = g_malloc(sizeof(gint));
    *key3 = 98; /* 98 Payload type is used for PS data */
    g_hash_table_insert(g_rtp_stream_info.rtp_dyn_payload,key3,"RTP_IP");

    return;
}

static int dissect_ursl_nasmsg(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree)
{
    call_dissector(nas_handle,tvb,pinfo,main_tree);

	return 0;
}

static int dissect_rrc_container(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree)
{
    if ((current_message_id == HANDOVERREQUEST_ID) ||
        (current_message_id == RELOCATIONREQUIRED_ID))
    {
        call_dissector(srnc_to_trnc_handle,tvb,pinfo,main_tree);
    }
    else if ((current_message_id == RELOCATIONCOMMAND_ID) || 
               (current_message_id == HANDOVERREQACK_ID ))
    {
        guint8  first_byte = tvb_get_guint8(tvb,0);
        if ((((first_byte & 0xf0)) == 0xc0) && (first_byte != 0xc5))
        {
            call_dissector(trnc_to_inter_rat_handle,tvb,pinfo,main_tree);
        }
        else
        {
            call_dissector(trnc_to_srnc_handle,tvb,pinfo,main_tree);
        }
    }
	return 0;
}

static int dissect_ursl_rtpserverinfo(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree)
{
    g_rtp_stream_info.packet_num = pinfo->fd->num;
    return 0;

}

static int dissect_ursl_acipaddr(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree)
{
    if (g_rtp_stream_info.packet_num == pinfo->fd->num)
    {
        g_rtp_stream_info.ac_rtp_ip_address[g_rtp_stream_info.curr_addr_index_no] =
                                        tvb_get_guint8(tvb,0);
        g_rtp_stream_info.curr_addr_index_no = (++g_rtp_stream_info.curr_addr_index_no) % 4;
    }
    return 0;
}
static int dissect_ursl_acudpport(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree)
{
    if (g_rtp_stream_info.packet_num == pinfo->fd->num)
    {
        g_rtp_stream_info.ac_rtp_port = tvb_get_ntohs(tvb,0);
        g_rtp_stream_info.ac_rtp_ip_addr.type = AT_IPv4;
        g_rtp_stream_info.ac_rtp_ip_addr.len = 4;
        g_rtp_stream_info.ac_rtp_ip_addr.data = se_alloc(4);
        memcpy((void *)g_rtp_stream_info.ac_rtp_ip_addr.data,
                    (void *)g_rtp_stream_info.ac_rtp_ip_address,4);

        g_rtp_stream_info.ac_rtcp_ip_addr.type = AT_IPv4;
        g_rtp_stream_info.ac_rtcp_ip_addr.len = 4;
        g_rtp_stream_info.ac_rtcp_ip_addr.data = se_alloc(4);
        memcpy((void *)g_rtp_stream_info.ac_rtcp_ip_addr.data,
                    (void *)g_rtp_stream_info.ac_rtp_ip_address,4);

        rtp_add_address(pinfo,&g_rtp_stream_info.ac_rtp_ip_addr,g_rtp_stream_info.ac_rtp_port,
                             0,"3GAP-RTP", pinfo->fd->num,g_rtp_stream_info.rtp_dyn_payload);
        rtcp_add_address(pinfo,&g_rtp_stream_info.ac_rtcp_ip_addr,g_rtp_stream_info.ac_rtp_port+1,
                             0,"3GAP-RTP", pinfo->fd->num);

    }
    return 0;
}
