/* packet-oyster-token_msg.c
* Register a dissector for ip.access's Token Message Format
* Copyright 2008 ip.access
*
* $Id: plugins/oyster3g/packet-oyster3g-token_msg.c 1.1 2008/08/05 13:34:07BST Norman Mackenzie (nm1) Exp  $
*
* Wireshark - Network traffic analyzer
* By Gerald Combs <gerald@wireshark.org>
* Copyright 1998 Gerald Combs
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "config.h"

#include <epan/packet.h>
#include <epan/prefs.h>
#include <epan/dissectors/packet-llc.h>
#include <epan/oui.h>
#include <epan/filesystem.h>
#include <wiretap/file_util.h>
#include <string.h>
#include <ctype.h>

#include "packet-oyster3g.h"

#define TOKENS_FILENAME             "tokens.txt"
#define FILTER_PREFIX               "ipa_tp.msg."
#define SUB_DISSECT_BUF_PREFIX      "dissect:"

/* Support for UDP echo of FPGA trace messages on a split system */
#define UDP_IPA_TOKEN_MESSAGE_PORT  13111

#define MAX_MESSAGE_BUFFER_BYTES    512

/* There are three types of argument to trace macros */
typedef enum message_arg_type_tag {
	E_INTEGER_ARG,
	E_STRTOK_ARG,
	E_STRING_ARG
} message_arg_type;

/* We'll read the tokens from a file into a linked list of this struct */
typedef struct token_definition_tag token_definition;

struct token_definition_tag {
    guint32            token;                   // Token num
    char*              string_p;                // Equivalent text
    char*              filename_p;              // Source filename
    gboolean           allows_string_literals;  // Whether this accepts string args
    gboolean           bufferMessage;           // Token is a buffer message
    dissector_handle_t subdissector;            // Sub-dissector for a buffer message
	guint8             num_args_defs;           // How many custom filters are present
	int*               headerId[32];            // A pointer per arg for custom filters
    token_definition*  next_p;
};

/* We'll try to construct custom argument filters using
   the info that's in the format string and the arg info
   for the token */
typedef struct message_arg_info_tag message_arg_info;

struct message_arg_info_tag {
    char*              name_p;         // Name to display for arg
    char*              filter_name_p;  // Name to use as a filter
    message_arg_type   type;           // Type (int, tok, str)
    int                id;             // The filter ID, this will be filled in by WS
	message_arg_info*  next_p;
};

/* If we can't read any tokens then use this empty array of strings */
static const value_string empty_token_string_vals[] = { 
    { 0, NULL }
};

static token_definition*     token_string_defs = NULL;
static const value_string*   token_string_vals = empty_token_string_vals;

static message_arg_info*     message_arg_defs = NULL;


/* We export three dissectors for clients. */

/* The first is for dissecting the FPGA Debug Card tokenised
   message format.  This has the app name, the format token,
   any formatted args and an optional data buffer. */
static int proto_ipatokmsg = -1;
static dissector_handle_t ipatokmsg_handle;

#define PROTO_SHORT_NAME "IPA_TOKMSG"
#define PROTO_LONG_NAME "ip.access Tokenised Message Protocol"

/* The second exported dissector is for displaying the
   text for string tokens in other protocol dissectors. */
static int proto_ipastrtok = -1;

#define STRTOK_PROTO_SHORT_NAME "IPA_STRTOK"
#define STRTOK_PROTO_LONG_NAME "ip.access String Token"

/* The final exported dissector is for displaying the
   text for appname tokens in other protocol dissectors. */
static int proto_ipaapptok = -1;

#define APPTOK_PROTO_SHORT_NAME "IPA_APPTOK"
#define APPTOK_PROTO_LONG_NAME "ip.access Application Name Token"



/* Standard filter IDs.  There are a number of filters for
   specific fields and arrays of filters for arguments of
   each type that will be used if there is no custom filter
   for the arg. */
static int hf_ipatokmsg_appname_token = -1;
static int hf_ipatokmsg_appname_string = -1;
static int hf_ipatokmsg_format_token = -1;
static int hf_ipatokmsg_format_string = -1;
static int hf_ipatokmsg_string_token = -1;
static int hf_ipatokmsg_string_mask = -1;
static int hf_ipatokmsg_width = -1;
static int hf_ipatokmsg_precision = -1;

static int hf_ipatokmsg_int_arg[32] = {
    -1, -1, -1, -1, -1, -1, -1, -1,
    -1, -1, -1, -1, -1, -1, -1, -1,
    -1, -1, -1, -1, -1, -1, -1, -1,
    -1, -1, -1, -1, -1, -1, -1, -1
};

static int hf_ipatokmsg_str_arg[32] = {
    -1, -1, -1, -1, -1, -1, -1, -1,
    -1, -1, -1, -1, -1, -1, -1, -1,
    -1, -1, -1, -1, -1, -1, -1, -1,
    -1, -1, -1, -1, -1, -1, -1, -1
};

static int hf_ipatokmsg_token_arg[32] = {
    -1, -1, -1, -1, -1, -1, -1, -1,
    -1, -1, -1, -1, -1, -1, -1, -1,
    -1, -1, -1, -1, -1, -1, -1, -1,
    -1, -1, -1, -1, -1, -1, -1, -1
};

/* These macros allow a shorthand defintion of the default filters
   for arguments */
#define IPA_TP_HEADER_FORMAT_FOR_ARG(x)\
{ &hf_ipatokmsg_int_arg[x-1],\
{ "Arg " #x " (int)", FILTER_PREFIX "intarg" #x, FT_UINT32, BASE_HEX, NULL, \
    0x0, "", HFILL }}, \
    \
{ &hf_ipatokmsg_str_arg[x-1],\
{ "Arg " #x " (str)", FILTER_PREFIX "strarg" #x, FT_STRINGZ, BASE_NONE, NULL,\
    0x0, "", HFILL }},\
    \
{ &hf_ipatokmsg_token_arg[x-1],\
{ "Arg " #x " (tok)", FILTER_PREFIX "tokarg" #x, FT_UINT32, BASE_HEX, NULL,\
    0x0, "", HFILL }}

/* Flag for recording expanded state of trace protocol messages */
static gint ett_ipatokmsg = -1;


/* Token lookup */
static token_definition*  find_trace_token(guint32 token)
{
    token_definition*   td = token_string_defs;

    while (td) {

        if (td->token == token) {

            break;
        }

        td = td->next_p;
    }

    return td;
}

/* Token to text/appname dissector for other protocols to use */
static int dissect_token_to_string(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree, int formatter)
{
    guint32             token;

    if (tvb_length_remaining(tvb, 0) != 4)
    {
        return 0;
    }

    token = tvb_get_letohl(tvb, 0);

    proto_tree_add_item(main_tree, formatter, tvb, 0, 4, FALSE);

    return 0;
}

/* Exported tokenised app name dissector */
static void
dissect_ipaapptok(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
    dissect_token_to_string(tvb, pinfo, tree, hf_ipatokmsg_appname_token);
}

/* Exported tokenised string dissector */
static void
dissect_ipastrtok(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
    dissect_token_to_string(tvb, pinfo, tree, hf_ipatokmsg_string_token);
}

/* Main FPGA tokenised message dissector */
static void
dissect_ipatokmsg(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
    int               offset=0;
    int               msg_len;
    proto_item        *msg_item=NULL;
    proto_tree        *msg_tree=NULL;
    token_definition  *msg_td = NULL;

    msg_len = tvb_length_remaining(tvb, offset);

    msg_item = proto_tree_add_text(tree, tvb, 0, msg_len, "Message: ");
    msg_tree = proto_item_add_subtree(msg_item, ett_ipatokmsg);

    if (msg_len >= 4)
    {
        char*              message = NULL;
        char*              message_p = NULL;
        int                message_space = 0;
        guint32            app_token;
        char*              appname_str = NULL;
        char*              filename_str = NULL;
        guint32            msg_token;
        char*              format_str = NULL;
        char               c;
        const char*        p = NULL;
        int                num_formats = 0;
        int                num_args = 0;
        guint32            string_mask;

        message = ep_alloc(MAX_MESSAGE_BUFFER_BYTES);
        if (!message) {
            return;
        }
        message[0] = 0;
        message_p = message;
        message_space = MAX_MESSAGE_BUFFER_BYTES-1;

        /* Read appname token.  If it's 0 then the appname follows as text,
           otherwise the token can be expanded to the appname */
        app_token = tvb_get_letohl(tvb, offset);
        proto_tree_add_item(msg_tree, hf_ipatokmsg_appname_token, tvb, offset, 4, TRUE);
        offset += 4;
        msg_len -= 4;

        if (app_token == 0) {

            /* Special case where application name is passed as a literal */
            guint32 appname_len;
            appname_str = tvb_get_ephemeral_stringz(tvb, offset, &appname_len);
            if (!appname_str) {
                return;
            }

            proto_tree_add_item(msg_tree, hf_ipatokmsg_appname_string, tvb, offset, appname_len, FALSE);
            appname_len += 1;
            appname_len &= ~1;

            offset += appname_len;
            msg_len -= appname_len;

        } else {

            token_definition*   td = find_trace_token(app_token);

            if (td && td->string_p) {
                appname_str = td->string_p;
            } else {
                appname_str = "";
            }
        }

        /* Read the format token.  If it's 0 then the format follows as text,
           otherwise the token can be expanded to the format */
        msg_token = tvb_get_letohl(tvb, offset);
        proto_tree_add_item(msg_tree, hf_ipatokmsg_format_token, tvb, offset, 4, TRUE);
        offset += 4;
        msg_len -= 4;

        if (msg_token == 0) {

            /* Special case where format string is passed as a literal */
            guint32 format_len;
            format_str = tvb_get_ephemeral_stringz(tvb, offset, &format_len);
            if (!format_str) {
                return;
            }

            proto_tree_add_item(msg_tree, hf_ipatokmsg_format_string, tvb, offset, format_len, FALSE);
            format_len += 1;
            format_len &= ~1;

            offset += format_len;
            msg_len -= format_len;

            /* If the format string isn't tokenised then neither are any args */
            string_mask = 0xFFFFFFFF;

            p = format_str;
            c = *p++;

        } else {

            msg_td = find_trace_token(msg_token);

            string_mask = 0;

            c = 0;

            if (msg_td) {

                if (msg_td->allows_string_literals) {

                    /* Macros that pass string literals will send us a string mask
                       so we can tell which are string vars and which are string tokens */
                    string_mask = tvb_get_letohl(tvb, offset);
                    proto_tree_add_item(msg_tree, hf_ipatokmsg_string_mask, tvb, offset, 4, TRUE);
                    offset += 4;
                    msg_len -= 4;
                }

                p = msg_td->string_p;
                if (p) {
                    c = *p++;
                } else {
                    c = 0;
                }

                if (msg_td->filename_p) {
                    filename_str = msg_td->filename_p;
                }
            }
        }

        /* Parse the format string, looking for specifiers. */
        /* Copy each format specifier into a string and use it to print the arg */
        while (c && (message_space > 0) && (num_args < 32) && (num_formats < 32)) {

            gint len_written = 0;

            if (c == '%') {

                c = *p++;

                if (c == '%') {

                    /* "%%" => '%' */
                    len_written = g_snprintf(message_p, message_space, "%%");

                } else {

                    int    star_fields = 0;
                    char   specifier[34];
                    char*  specifier_p = specifier;
                    int    specifier_len = 0;

                    *specifier_p++ = '%';
                    specifier_len++;

                    /* Flags */
                    while (strchr("-+ #0", c) && (specifier_len < 32)) {
                        *specifier_p++ = c;
                        specifier_len++;
                        c = *p++;
                    }
                    /* Width */
                    if (c == '*') {
                        if (specifier_len < 32) {
                            *specifier_p++ = c;
                        }
                        c = *p++;
                        star_fields |= 1;
                    } else {
                        while (isdigit(c) && (specifier_len < 32)) {
                            *specifier_p++ = c;
                            c = *p++;
                        }
                    }
                    /* Precision */
                    if (c == '.') {
                        if (specifier_len < 32) {
                            *specifier_p++ = c;
                        }
                        c = *p++;
                        if (c == '*') {
                            if (specifier_len < 32) {
                                *specifier_p++ = c;
                            }
                            c = *p++;
                            star_fields |= 2;
                        } else {
                            while (isdigit(c) && (specifier_len < 32)) {
                                *specifier_p++ = c;
                                c = *p++;
                            }
                        }
                    }
                    /* Length modifier */
                    if (strchr("jztL", c)) {
                        if (specifier_len < 32) {
                            *specifier_p++ = c;
                        }
                        c = *p++;
                    } else if (c == 'h') {
                        if (specifier_len < 32) {
                            *specifier_p++ = c;
                        }
                        c = *p++;
                        if (c == 'h') {
                            if (specifier_len < 32) {
                                *specifier_p++ = c;
                            }
                            c = *p++;
                        }
                    } else if (c == 'l') {
                        if (specifier_len < 32) {
                            *specifier_p++ = c;
                        }
                        c = *p++;
                        if (c == 'l') {
                            if (specifier_len < 32) {
                                *specifier_p++ = c;
                            }
                            c = *p++;
                        }
                    }

                    /* Conversion specifier */
                    if (specifier_len < 32) {

                        int      custom_filter = -1;
                        guint32  width         = 0;
                        guint32  precision     = 0;

                        /* Is there a custom filter associated with this argument */
                        if (msg_td) {
                            if (num_formats < msg_td->num_args_defs) {
                                if (msg_td->headerId[num_formats]) {
                                    custom_filter = *msg_td->headerId[num_formats];
                                }
                            }
                        }

                        *specifier_p++ = c;
                        *specifier_p++ = 0;

                        if ((star_fields & 1) && (msg_len >= 4)) {
                            width = tvb_get_letohl(tvb, offset);
                            proto_tree_add_item(msg_tree, hf_ipatokmsg_width, tvb, offset, 4, TRUE);
                            offset += 4;
                            msg_len -= 4;
                            num_args++;
                        }
                        if ((star_fields & 2) && (msg_len >= 4)) {
                            precision = tvb_get_letohl(tvb, offset);
                            proto_tree_add_item(msg_tree, hf_ipatokmsg_precision, tvb, offset, 4, TRUE);
                            offset += 4;
                            msg_len -= 4;
                            num_args++;
                        }

                        if (strchr("dioucpxX", c)) {
                            /* Integer type */
                            if  (msg_len >= 4) {

                                guint32 int_val = tvb_get_letohl(tvb, offset);

                                if (custom_filter != -1) {
                                    proto_tree_add_item(msg_tree, custom_filter, tvb, offset, 4, TRUE);
                                } else {
                                    proto_tree_add_item(msg_tree, hf_ipatokmsg_int_arg[num_formats], tvb, offset, 4, TRUE);
                                }
                                offset += 4;
                                msg_len -= 4;

                                switch (star_fields) {
                                    case 3:
                                        len_written = g_snprintf(message_p, message_space, specifier, width, precision, int_val);
                                        break;
                                    case 2:
                                        len_written = g_snprintf(message_p, message_space, specifier, precision, int_val);
                                        break;
                                    case 1:
                                        len_written = g_snprintf(message_p, message_space, specifier, width, int_val);
                                        break;
                                    default:
                                        len_written = g_snprintf(message_p, message_space, specifier, int_val);
                                        break;
                                }

                            } else {

                                len_written = g_snprintf(message_p, message_space, "?");
                            }

                        } else if (c == 's') {

                            /* Read a string token or string literal */
                            const char*     str_arg = NULL;
                            const char*     src_p;
                            char            c;
                            int             escaped_chars = 0;

                            if (string_mask & (1<<num_args))
                            {
                                guint32   arg_len = 0;

                                str_arg = tvb_get_ephemeral_stringz(tvb, offset, &arg_len);
                                if (!str_arg) {
                                    return;
                                }

                                if (custom_filter != -1) {
                                    proto_tree_add_item(msg_tree, custom_filter, tvb, offset, arg_len, FALSE);
                                } else {
                                    proto_tree_add_item(msg_tree, hf_ipatokmsg_str_arg[num_formats], tvb, offset, arg_len, FALSE);
                                }

                                arg_len += 1;
                                arg_len &= ~1;

                                offset += arg_len;
                                msg_len -= arg_len;
                            }
                            else
                            {
                                guint32 string_token;

                                token_definition*   td = token_string_defs;

                                string_token = tvb_get_letohl(tvb, offset);
                                if (custom_filter != -1) {
                                    proto_tree_add_item(msg_tree, custom_filter, tvb, offset, 4, TRUE);
                                } else {
                                    proto_tree_add_item(msg_tree, hf_ipatokmsg_token_arg[num_formats], tvb, offset, 4, TRUE);
                                }
                                offset += 4;
                                msg_len -= 4;

                                str_arg = "?";

                                while (td) {

                                    if (td->token == string_token) {
                                        str_arg = td->string_p;
                                        break;
                                    }

                                    td = td->next_p;
                                }
                            }

                            /* Expand characters that are the result of escape sequences
                               back to the escape sequence and output all other non-printable
                               characters as a dot. */
                            for (src_p=str_arg; *src_p; src_p++)
                            {
                                c = *src_p;
                                if (strchr("\r\n\t\b\a\f\v", c))
                                {
                                    escaped_chars++;
                                }
                            }

                            if (escaped_chars)
                            {
                                char*  expanded_str_arg = NULL;
                                char*  dst_p;

                                expanded_str_arg = ep_alloc(escaped_chars + strlen(str_arg) + 1);
                                if (!expanded_str_arg) {
                                    return;
                                }

                                src_p = str_arg;
                                dst_p = expanded_str_arg;

                                do {

                                    c = *src_p++;
                                    if ((c == 0) || isprint(c)) {
                                        *dst_p++ = c;
                                    } else {
                                        char* escaped_char_p = strchr("\rr\nn\tt\bb\aa\ff\vv", c);

                                        if (escaped_char_p) {
                                            *dst_p++ = '\\';
                                            *dst_p++ = *(escaped_char_p+1);
                                        } else {
                                            *dst_p++ = '.';
                                        }
                                    }
                                } while (c);

                                str_arg = expanded_str_arg;
                            }

                            switch (star_fields) {
                                case 3:
                                    len_written = g_snprintf(message_p, message_space, specifier, width, precision, str_arg);
                                    break;
                                case 2:
                                    len_written = g_snprintf(message_p, message_space, specifier, precision, str_arg);
                                    break;
                                case 1:
                                    len_written = g_snprintf(message_p, message_space, specifier, width, str_arg);
                                    break;
                                default:
                                    len_written = g_snprintf(message_p, message_space, specifier, str_arg);
                                    break;
                            }
                        }

                        num_args++;
                        num_formats++;
                    }
                }

            } else if (c == '\n') {

                len_written = g_snprintf(message_p, message_space, "%c%c", '\\', 'n');

            } else if (c == '\r') {

                len_written = g_snprintf(message_p, message_space, "%c%c", '\\', 'r');

            } else {

                len_written = g_snprintf(message_p, message_space, "%c", c);

            }

            message_p += len_written;
            message_space -= len_written;

            c = *p++;
        }

        *message_p = 0;
        if (filename_str) {
            proto_item_append_text(msg_item, "File %s: ", filename_str);
        }
        if (msg_len > 0) {
            proto_item_append_text(msg_item, "\"%s:%s\" + %d bytes of data", appname_str, message, msg_len);
        } else {
            proto_item_append_text(msg_item, "%s:%s", appname_str, message);
        }

        if (check_col(pinfo->cinfo, COL_INFO)) {
            col_set_str(pinfo->cinfo, COL_INFO, "");
            if (msg_len > 0) {
                col_append_fstr(pinfo->cinfo, COL_INFO, "\"%s:%s\" + %d bytes of data", appname_str, message, msg_len);
            } else {
                col_append_fstr(pinfo->cinfo, COL_INFO, "%s:%s", appname_str, message);
            }
        }
    }

    if (msg_len > 0) {

        if (msg_td && msg_td->subdissector) {

            tvbuff_t *next_tvb = tvb_new_subset(tvb, offset, msg_len, -1);

            call_dissector(msg_td->subdissector, next_tvb, pinfo, msg_tree);

        } else {

            proto_tree_add_text(msg_tree, tvb, offset, msg_len, "Data buffer (%d bytes)", msg_len);
            offset += msg_len;
        }
    }
}

/* Find a supported conversion specifier in a format string, if the
   function returns NULL then none was found, otherwise it returns
   the point to continue the search for the next conversion specifier.
   Pass back whether the specifier is for a string (or integer/token)
   and also pass back a pointer to an identifier for the arg filter if
   appropriate.  That will be an identifier followed by =, immediately
   before the conversion specifier.  The identifier will be terminated
   in the format string (so the format string will be modified) and a
   pointer to that returned in filter, or NULL if no filter found.
 */
static char*
scan_arg_info(char* p, int* str, char** filter)
{
	char*  retVal  = NULL;
	char*  word    = NULL;
	char*  wordend = NULL;
	int    inword  = FALSE;
	int    equals  = FALSE;
	char   c;

	*filter = NULL;

	do {

		c = *p++;

		if (!inword && isalpha(c)) {

			inword = TRUE;
			word = p - 1;

		} else if (inword && !isalnum(c) && (c != '_')) {

			wordend = p - 1;
			inword = FALSE;

		}
		
		if ((c == '=') && word) {

			equals = TRUE;

		} else if (c == '%') {

			if (*p == '%') {

				p++;

			} else {

				break;
			}

		} else if (!isspace(c)) {

			equals = FALSE;
		}

	} while (c);

	if (c) {
		c = *p++;
	} else {
		return NULL;
	}

	/* Flags */
	while (strchr("-+ #0", c)) {
		c = *p++;
	}

	/* Width */
	if (c == '*') {
		c = *p++;
	} else {
		while (isdigit(c)) {
			c = *p++;
		}
	}

	/* Precision */
	if (c == '.') {
		c = *p++;
		if (c == '*') {
			c = *p++;
		} else {
			while (isdigit(c)) {
				c = *p++;
			}
		}
	}

	/* Length modifier */
	if (strchr("jztL", c)) {
		c = *p++;
	} else if (c == 'h') {
		c = *p++;
		if (c == 'h') {
			c = *p++;
		}
	} else if (c == 'l') {
		c = *p++;
		if (c == 'l') {
			c = *p++;
		}
	}

	/* Conversion specifier */
	if (strchr("dioucpxX", c)) {

		retVal = p;
		*str = FALSE;

	} else if (c == 's') {

		retVal = p;
		*str = TRUE;
	}

	if (retVal && equals) {

		*wordend = 0;
		*filter = word;
	}

	return retVal;
}

/* Skip white space in a token defintion line */
static void
tokens_file_skip_space(char** c_pp)
{
    char*  c_p = *c_pp;

    while (isspace(*c_p)) {
        c_p++;
    }

    *c_pp = c_p;
}

/* Skip white space in a token defintion line, ensuring that there is some */
static gboolean
tokens_file_check_and_skip_space(char** c_pp)
{
    char*  c_p = *c_pp;

    if (!isspace(*c_p)) {
        return FALSE;
    }

    while (isspace(*c_p)) {
        c_p++;
    }

    *c_pp = c_p;

    return TRUE;
}

/* Read a hex constant (token) from a token defintion line */
static gboolean
tokens_file_read_hex(char** c_pp, guint32* num_p)
{
    guint32 num = 0;
    char*   c_p = *c_pp;

    if (!isxdigit(*c_p)) {
        return FALSE;
    }

    while (isxdigit(*c_p)) {
        num <<= 4;
        if (isdigit(*c_p)) {
            num |= *c_p - '0';
        } else {
            num |= tolower(*c_p) - 'a' + 10;
        }
        c_p++;
    }

    *num_p = num;
    *c_pp  = c_p;

    return TRUE;
}

/* Read a non-whitespace field (e.g. filename) from a token defintion line */
static gboolean
tokens_file_find_field(char** c_pp, char** start_pp, char** end_pp)
{
    char*  c_p;

    /* Should have white space between previous field and this */
    if (!tokens_file_check_and_skip_space(c_pp)) {
        return FALSE;
    }
    c_p = *c_pp;

    /* Field is a sequence of one or more non-white space */
    *start_pp = c_p;

    if (0 == *c_p) {
        return FALSE;
    }

    while (*c_p && !isspace(*c_p)) {
        c_p++;
    }

    *end_pp = c_p;

    *c_pp = c_p;

    return TRUE;
}

/* Read a string from a token defintion line, allowing for embedded escaped chars */
static gboolean
tokens_file_find_string(char** c_pp, char** start_pp, char** end_pp)
{
    char*     c_p;
    gboolean  escape   = FALSE;
    gboolean  complete = FALSE;

    /* Should have white space between previous field and this */
    if (!tokens_file_check_and_skip_space(c_pp)) {
        return FALSE;
    }
    c_p = *c_pp;

    /* Field is a quoted field, with zero or more enclosed escaped quotes */
    if ('"' != *c_p) {
        return FALSE;
    }
    c_p++;

    /* Don't include the quotes in the result */
    *start_pp = c_p;

    while (*c_p && !complete) {
        if (escape) {
            escape = FALSE;
        } else {
            if (*c_p == '\\') {
                escape = TRUE;
            } else if (*c_p == '"') {
                *end_pp = c_p;
                complete = TRUE;
            }
        }
        c_p++;
    }

    *c_pp = c_p;

    return complete;
}

/* Read a macro argument from a token defintion line.  Allow for quoted string etc. */
static gboolean
tokens_file_find_arg(char** c_pp, char** start_pp, char** end_pp)
{
    char*     c_p;
    int       paren_level = 0;
    gboolean  in_string_lit = FALSE;
    gboolean  in_char_lit   = FALSE;
    gboolean  in_escape_seq = FALSE;

    /* Should have white space between previous field and this */
    if (!tokens_file_check_and_skip_space(c_pp)) {
        return FALSE;
    }
    c_p = *c_pp;

    /* Field is a sequence of one or more non-white space with the
	   possibility of embedded strings and chars */
    *start_pp = c_p;

    if (0 == *c_p) {
        return FALSE;
    }

    while (*c_p) {
        if (in_string_lit) {
            if (!in_escape_seq) {
                if (*c_p == '\\') {
                    in_escape_seq = TRUE;
                } else if (*c_p == '"') {
                    in_string_lit = FALSE;
                }
            } else {
                in_escape_seq = FALSE;
            }
        } else if (in_char_lit) {
            if (!in_escape_seq) {
                if (*c_p == '\\') {
                    in_escape_seq = TRUE;
                } else if (*c_p == '\'') {
                    in_char_lit = FALSE;
                }
            } else {
                in_escape_seq = FALSE;
            }
        } else if (*c_p == '"') {
            in_string_lit = TRUE;
        } else if (*c_p == '\'') {
            in_char_lit = TRUE;
        } else if (*c_p == '(') {
            paren_level++;
        } else if (*c_p == ')') {
            paren_level--;
        } else if (isspace(*c_p) && (paren_level==0)) {
            break;
        }
        c_p++;
    }

    if ((paren_level > 0) || in_escape_seq || in_char_lit || in_string_lit) {
        return FALSE;
    }

    *end_pp = c_p;

    *c_pp = c_p;

    return TRUE;
}

/* Add a custom filter defintion to their linked list.
   Return a pointer to the id used for the filter or NULL if
   there was a problem adding the filter.  Ensure that we only
   have a single defintion for each filter.
 */
int *
create_filter(char* s, int len, message_arg_type type)
{
    char*             new_text_p        = NULL;
    char*             new_filter_text_p = NULL;
    message_arg_info* new_filter_p      = NULL;
    message_arg_info* filter_p          = NULL;
    
    new_filter_p = (message_arg_info*)g_malloc(sizeof(message_arg_info));
    if (!new_filter_p) {
        return NULL;
    }

    new_text_p = (char*)g_malloc(len+1);
    if (!new_filter_p) {
        g_free(new_filter_p);
        return NULL;
    }

    new_filter_text_p = (char*)g_malloc(len+1+strlen(FILTER_PREFIX));
    if (!new_filter_p) {
        g_free(new_filter_p);
        g_free(new_text_p);
        return NULL;
    }

    memcpy(new_text_p, s, len);
    new_text_p[len] = 0;

    strcpy(new_filter_text_p, FILTER_PREFIX);
    memcpy(new_filter_text_p + strlen(FILTER_PREFIX), s, len);
    new_filter_text_p[len + strlen(FILTER_PREFIX)] = 0;

    new_filter_p->name_p        = new_text_p;
    new_filter_p->filter_name_p = new_filter_text_p;
    new_filter_p->type          = type;
    new_filter_p->id            = -1;

    for (filter_p=message_arg_defs; filter_p; filter_p=filter_p->next_p) {
        if (filter_p->type == new_filter_p->type) {
            if (0 == strcmp(filter_p->name_p, new_filter_p->name_p)) {
                g_free(new_text_p);
                g_free(new_filter_p);
                return &(filter_p->id);
            }
        }
    }

    new_filter_p->next_p = message_arg_defs;
    message_arg_defs = new_filter_p;
    return &(new_filter_p->id);
}

/* Scan a line from the tokens definitions file */
static gboolean
tokens_file_parse_line(char* c_p)
{
    guint32            token_num      = 0;
    guint32            arg_num        = 0;

    char*              filename_p     = NULL;
    char*              filename_end_p = NULL;
    int                filename_len   = 0;
    char*              new_filename_p = NULL;

    char*              version_p      = NULL;
    char*              version_end_p  = NULL;

    char*              type_p         = NULL;
    char*              type_end_p     = NULL;

    char*              format_p       = NULL;
    char*              format_end_p   = NULL;
    int                format_len     = 0;
    char*              new_format_p   = NULL;

    char*              arg_p          = NULL;
    char*              arg_end_p      = NULL;
	char*              filter_p       = NULL;
	gboolean           str_arg;

    token_definition*  td             = NULL;

    tokens_file_skip_space(&c_p);

    /* Ignore blank lines and comment lines */
    if ((*c_p == '#') || (*c_p == 0)) {
        return TRUE;
    }

    /* Read the hex token number */
    if (!tokens_file_read_hex(&c_p, &token_num)) {
        return FALSE;
    }

    /* Find the filename */
    if (!tokens_file_find_field(&c_p, &filename_p, &filename_end_p)) {
        return FALSE;
    }

    /* Find the version */
    if (!tokens_file_find_field(&c_p, &version_p, &version_end_p)) {
        return FALSE;
    }

    /* Find the type field */
    if (!tokens_file_find_field(&c_p, &type_p, &type_end_p)) {
        return FALSE;
    }

    /* Find a format string */
    if (!tokens_file_find_string(&c_p, &format_p, &format_end_p)) {
        return FALSE;
    }

    /* Terminate the separate fields */
    *filename_end_p = 0;
    *version_end_p  = 0;
    *type_end_p     = 0;
    *format_end_p   = 0;

    format_len = format_end_p - format_p;
    filename_len = filename_end_p - filename_p;

    /* Have enough info for a definition.  Create a new one. */
    td = (token_definition*)g_malloc(sizeof(token_definition));
    if (!td) {
        return FALSE;
    }

    if ((0 == strcmp(type_p, "DEBUG")) || (0 == strcmp(type_p, "TRACE"))) {
        td->allows_string_literals = TRUE;
    } else {
        td->allows_string_literals = FALSE;
    }
    td->token = token_num;

    td->subdissector = NULL;
    td->bufferMessage = (0 == strcmp(type_p, "BUFFER"));

    /* Get a copy of the format string */
    new_format_p = (char*)g_malloc(format_len + 2);
    if (!new_format_p) {
        g_free(td);
        return FALSE;
    }

    g_snprintf(new_format_p, format_len + 1, "%s", format_p);
    td->string_p = new_format_p;

    /* Get a copy of the filename */
    new_filename_p = (char*)g_malloc(filename_len + 2);
    if (!new_filename_p) {
        g_free(new_format_p);
        g_free(td);
        return FALSE;
    }

    g_snprintf(new_filename_p, filename_len + 1, "%s", filename_p);
    td->filename_p = new_filename_p;

    /* Scan the arguments for custom filters we may be able to construct */
    td->num_args_defs = 0;

	while (NULL != (format_p = scan_arg_info(format_p, &str_arg, &filter_p))) {

		if (tokens_file_find_arg(&c_p, &arg_p, &arg_end_p)) {

			int   token = 0 == strncmp(arg_p, "STRING_TOK(", strlen("STRING_TOK("));
			int   type;
            int*  id_p = NULL;

			if (str_arg && token) {
				type = E_STRTOK_ARG;
			} else if (str_arg) {
				type = E_STRING_ARG;
			} else {
				type = E_INTEGER_ARG;
			}

			if (filter_p) {

                /* We have something like "filter_name = %d" in the format string */
				id_p = create_filter(filter_p, strlen(filter_p), type);

			} else {

                /* Check if the macro was passed an arg that could be used as a filter
                   For example, p->msg_type can provide a filter of msg_type.
                 */
				char*     f_p;
                char*     arg_start = NULL;

				for (f_p=arg_p; f_p<arg_end_p; f_p++) {
                    if (arg_start) {
                        if (!isalnum(*f_p) && (*f_p != '_')) {
                            arg_start = NULL;
                        }
                    } else {
                        if (isalpha(*f_p)) {
                            arg_start = f_p;
                        }
                    }
				}

				if (arg_start) {
					id_p = create_filter(arg_start, arg_end_p-arg_start, type);
				}
			}

            /* Save the pointer to the filter ID for the arg in the token info */
            td->headerId[arg_num] = id_p;

        	arg_num++;

            td->num_args_defs = arg_num;
		}
        else
        {
            /* Mismatch between the format string and the args. */
            break;
        }
	}

    /* Add definition to the linked list */
    td->next_p = token_string_defs;
    token_string_defs = td;

    return TRUE;
}

void
proto_register_ipatokmsg(void)
{
    /* Standard filters, without any custom additions */
    static hf_register_info hf[] = {

        /* IPA_TP tokenised fields are defined here */
        { &hf_ipatokmsg_appname_token,
        { "Application Name Token", "ipa_tp.msg.appname_token", FT_UINT32, BASE_HEX, NULL,
        0x0, "", HFILL }},

        { &hf_ipatokmsg_format_token,
        { "Format Token", "ipa_tp.msg.format_token", FT_UINT32, BASE_HEX, NULL,
        0x0, "", HFILL }},

        { &hf_ipatokmsg_string_token,
        { "String Token", "ipa_tp.msg.string_token", FT_UINT32, BASE_HEX, NULL,
        0x0, "", HFILL }},

        IPA_TP_HEADER_FORMAT_FOR_ARG(1),
        IPA_TP_HEADER_FORMAT_FOR_ARG(2),
        IPA_TP_HEADER_FORMAT_FOR_ARG(3),
        IPA_TP_HEADER_FORMAT_FOR_ARG(4),
        IPA_TP_HEADER_FORMAT_FOR_ARG(5),
        IPA_TP_HEADER_FORMAT_FOR_ARG(6),
        IPA_TP_HEADER_FORMAT_FOR_ARG(7),
        IPA_TP_HEADER_FORMAT_FOR_ARG(8),
        IPA_TP_HEADER_FORMAT_FOR_ARG(9),
        IPA_TP_HEADER_FORMAT_FOR_ARG(10),
        IPA_TP_HEADER_FORMAT_FOR_ARG(11),
        IPA_TP_HEADER_FORMAT_FOR_ARG(12),
        IPA_TP_HEADER_FORMAT_FOR_ARG(13),
        IPA_TP_HEADER_FORMAT_FOR_ARG(14),
        IPA_TP_HEADER_FORMAT_FOR_ARG(15),
        IPA_TP_HEADER_FORMAT_FOR_ARG(16),
        IPA_TP_HEADER_FORMAT_FOR_ARG(17),
        IPA_TP_HEADER_FORMAT_FOR_ARG(18),
        IPA_TP_HEADER_FORMAT_FOR_ARG(19),
        IPA_TP_HEADER_FORMAT_FOR_ARG(20),
        IPA_TP_HEADER_FORMAT_FOR_ARG(21),
        IPA_TP_HEADER_FORMAT_FOR_ARG(22),
        IPA_TP_HEADER_FORMAT_FOR_ARG(23),
        IPA_TP_HEADER_FORMAT_FOR_ARG(24),
        IPA_TP_HEADER_FORMAT_FOR_ARG(25),
        IPA_TP_HEADER_FORMAT_FOR_ARG(26),
        IPA_TP_HEADER_FORMAT_FOR_ARG(27),
        IPA_TP_HEADER_FORMAT_FOR_ARG(28),
        IPA_TP_HEADER_FORMAT_FOR_ARG(29),
        IPA_TP_HEADER_FORMAT_FOR_ARG(30),
        IPA_TP_HEADER_FORMAT_FOR_ARG(31),
        IPA_TP_HEADER_FORMAT_FOR_ARG(32),

        /* Trace message */
        { &hf_ipatokmsg_appname_string,
        { "Application Name String", "ipa_tp.msg.appname_string", FT_STRINGZ, BASE_NONE, NULL,
        0x0, "", HFILL }},

        { &hf_ipatokmsg_format_string,
        { "Format String", "ipa_tp.msg.format_string", FT_STRINGZ, BASE_NONE, NULL,
        0x0, "", HFILL }},

        { &hf_ipatokmsg_string_mask,
        { "String Arg Mask", "ipa_tp.msg.string_mask", FT_UINT32, BASE_HEX, NULL,
        0x0, "", HFILL }},

        { &hf_ipatokmsg_width,
        { "Field width", "ipa_tp.msg.field_width", FT_UINT32, BASE_DEC, NULL,
        0x0, "", HFILL }},

        { &hf_ipatokmsg_precision,
        { "Field precision", "ipa_tp.msg.field_precision", FT_UINT32, BASE_DEC, NULL,
        0x0, "", HFILL }},

    };

    static gint *ett[] = {
        &ett_ipatokmsg,
    };

    hf_register_info*   filters_p         = hf;
    int                 number_of_filters = array_length(hf);

    /* Read the token definitions */
    char*    tokens_filename = NULL;

    /* Try to use a tokens file from the user's personal configuration
       area.  If the isn't one then try the global area */
	tokens_filename = get_persconffile_path(TOKENS_FILENAME, TRUE, FALSE);
    if (tokens_filename) {
        if (!file_exists(tokens_filename)) {
            g_free(tokens_filename);
            tokens_filename = get_datafile_path(TOKENS_FILENAME);
        }
    }

    if (tokens_filename && file_exists(tokens_filename)) {

        FILE*  tokens_file = eth_fopen(tokens_filename, "r");

        if (tokens_file) {

            char*  input_line = g_malloc(MAX_MESSAGE_BUFFER_BYTES+1);

            if (input_line) {

                while (fgets(input_line, MAX_MESSAGE_BUFFER_BYTES, tokens_file)) {

                    input_line[MAX_MESSAGE_BUFFER_BYTES] = 0;

                    if (!tokens_file_parse_line(input_line)) {
                        break;
                    }
                }

                g_free(input_line);
            }

            fclose(tokens_file);
        }
    }

    if (tokens_filename) {
        g_free(tokens_filename);
    }

    /* If we have token definitions then we want to create a value_string array
        that allows the dissector to display the equivalent text for each token
     */
    if (token_string_defs) {

        guint32             number_of_tokens = 1; // Need at least a terminator
        token_definition*   td = token_string_defs;
        value_string*       string_vals = NULL;

        /* Count the number of tokens. */
        while (td) {
            number_of_tokens++;
            td = td->next_p;
        }

        /* Allocate enough space in a string array for them all.  Allow for
           a special case of 0 = > literal text. */
        string_vals = g_new(value_string, number_of_tokens+1);

        if (string_vals) {

            /* Copy the tokens and the equivalent texts into the array */
            int  i;

            token_string_vals = string_vals;

            td = token_string_defs;

            string_vals->value  = 0;
            string_vals->strptr = "(literal)";
            string_vals++;
            number_of_tokens--;

            while (td && number_of_tokens) {
                string_vals->value  = td->token;
                string_vals->strptr = td->string_p;
                string_vals++;
                number_of_tokens--;
                td = td->next_p;
            }

            string_vals->value  = 0;
            string_vals->strptr = NULL;

            /* Then point the types that use string tokens to this array */
            hf[0].hfinfo.strings = token_string_vals;
            hf[1].hfinfo.strings = token_string_vals;
            hf[2].hfinfo.strings = token_string_vals;

            for (i=0; i<32; ++i) {
                hf[3*i+5].hfinfo.strings = token_string_vals;
            }
        }
    }

    /* Build any custom filters that we got from the tokens file */
    if (message_arg_defs) {

        message_arg_info*   fd = message_arg_defs;
        hf_register_info    empty = { NULL, { "", "", FT_UINT32, BASE_HEX, NULL, 0x0, "", HFILL }};

        /* Count the filters */
        number_of_filters = 0;

        while (fd) {
            number_of_filters++;
            fd = fd->next_p;
        }

        /* Allocate space for an array of these filters plus the standard ones */
        filters_p = g_new(hf_register_info, number_of_filters+array_length(hf));

        if (filters_p) {

            hf_register_info* dest = filters_p;

            /* Copy the standard (default) filters into the start of this array */
            memcpy(filters_p, hf, sizeof(hf));

            /* Copy the custom ones after that */
            dest += array_length(hf);

            for (fd=message_arg_defs; fd; fd=fd->next_p) {

                memcpy(dest, &empty, sizeof(hf_register_info));

                dest->p_id          = &(fd->id);
                dest->hfinfo.name   = fd->name_p;
                dest->hfinfo.abbrev = fd->filter_name_p;

                if (fd->type == E_STRTOK_ARG) {
                    dest->hfinfo.strings = token_string_vals;
                } else if (fd->type == E_STRING_ARG) {
                    dest->hfinfo.type = FT_STRINGZ;
                    dest->hfinfo.display = BASE_NONE;
                }

                dest++;
            }

            number_of_filters += array_length(hf);

        } else {

            /* Failed to create custom filters.  Just use the standard ones */
            filters_p = hf;
            number_of_filters = array_length(hf);
        }
    }

    proto_ipatokmsg = proto_register_protocol(PROTO_LONG_NAME, PROTO_SHORT_NAME, "ipa_tokmsg");
    proto_ipastrtok = proto_register_protocol(STRTOK_PROTO_LONG_NAME, STRTOK_PROTO_SHORT_NAME, "ipa_strtok");
    proto_ipaapptok = proto_register_protocol(APPTOK_PROTO_LONG_NAME, APPTOK_PROTO_SHORT_NAME, "ipa_apptok");

    register_dissector("ipa_tokmsg", dissect_ipatokmsg, proto_ipatokmsg);
    register_dissector("ipa_strtok", dissect_ipastrtok, proto_ipastrtok);
    register_dissector("ipa_apptok", dissect_ipaapptok, proto_ipaapptok);

    proto_register_field_array(proto_ipatokmsg, filters_p, number_of_filters);
    proto_register_subtree_array(ett, array_length(ett));
}

void
proto_reg_handoff_ipatokmsg(void)
{
  token_definition* td;

  ipatokmsg_handle = find_dissector("ipa_tokmsg");

  /* If there are any buffer type tokens that expand to text that includes
     the SUB_DISSECT_BUF_PREFIX, then treat the rest of the text as a sub-dissector
     to call to display the contents of the buffer */
  for (td=token_string_defs; td; td=td->next_p) {
    if (td->bufferMessage && td->string_p) {
      char* startOfSubDis = strstr(td->string_p, SUB_DISSECT_BUF_PREFIX);
      if (startOfSubDis) {
        startOfSubDis += strlen(SUB_DISSECT_BUF_PREFIX);
        if (strlen(startOfSubDis) > 0) {
          td->subdissector = find_dissector(startOfSubDis);
        }
      }
    }
  }

#ifdef UDP_IPA_TOKEN_MESSAGE_PORT
  /* Support for UDP echo packets containing tokenised messages */
  dissector_add("udp.port", UDP_IPA_TOKEN_MESSAGE_PORT, ipatokmsg_handle);
#endif
}
