/* packet-oyster3g.c
 * Routines for ip.access oyster3G Signalling over IP framing dissection
 * Copyright 2007, ip.access Ltd <support@ipaccess.com>
 *
 * $Id: plugins/oyster3g/packet-oyster3g.c 1.20 2008/02/18 13:50:44GMT Neil P Piercy (npp) Exp  $
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@unicom.net>
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdlib.h>

#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#include <glib.h>

#ifdef NEED_SNPRINTF_H
# include "snprintf.h"
#endif

#include <epan/emem.h>
#include <epan/packet.h>
#include <epan/expert.h>
#include <epan/conversation.h>

#include "packet-oyster3g.h"
#include <epan/dissectors/packet-ssl.h>

/* function prototypes for this file */
static int dissect_oyster3g(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree);
static int dissect_oyster3g_frame(tvbuff_t *tvb, int offset, packet_info *pinfo, proto_tree *tree, int *desegment);

/* Initialize the protocol and registered fields for the SoIP frames */
static int proto_oyster3g = -1;

static int hf_msg_len = -1;
static int hf_stream = -1;
static int hf_version = -1;
static int hf_segmentation = -1;

static dissector_table_t subdissector_table;

static dissector_handle_t oyster3g_handle;
static dissector_handle_t data_handle;

/* Initialize the subtree pointers */
static gint ett_oyster3g = -1;

/* Value Strings */
static const value_string segmentation_vals[]= {
    { 0, "Complete Message" },
    { 1, "First Segment" },
    { 2, "Middle Segment" },
    { 3, "Last Segment" },
    { 0, NULL }
};
#define OYSTER3G_SOIP_SEGMENTATION_LAST  3

static gboolean dissect_oyster3g_heur(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
    gint offset=0;
    guint16 oyster3g_len;
    guint8 stream;
    guint8 version;
    guint8 segmentation;
    gboolean rc=FALSE;
    conversation_t *conversation;

    if (tvb_length(tvb)>=4) {
        oyster3g_len = tvb_get_ntohs(tvb, offset);
        offset += 2;
        stream = tvb_get_guint8(tvb, offset);
        offset++;
        version = tvb_get_guint8(tvb, offset);
        offset++;
        segmentation = version & 0x0f;
        version = version >> 4;

        if ((version == OYSTER3G_SOIP_VERSION)
         && (segmentation <= OYSTER3G_SOIP_SEGMENTATION_LAST)
         && (oyster3g_len==tvb_length_remaining(tvb,offset)))
            rc=TRUE;
    }

    if (rc==TRUE) {
        conversation = find_conversation(pinfo->fd->num, &pinfo->src, &pinfo->dst, pinfo->ptype, pinfo->srcport, pinfo->destport, 0);
        if (!conversation)
            conversation = conversation_new(pinfo->fd->num, &pinfo->src, &pinfo->dst, pinfo->ptype, pinfo->srcport, pinfo->destport, 0);
        conversation_set_dissector(conversation, oyster3g_handle);
        dissect_oyster3g(tvb, pinfo, tree);
    }

    return rc;
}

/* Code to actually dissect the packets - actually only a stub to allow the real
   routine (which this calls) to be called multiple times for packets which contain more than one message
*/
static int dissect_oyster3g(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
    int offset=0, this_msg_len, desegment;

    if (check_col(pinfo->cinfo, COL_PROTOCOL))
        col_add_str(pinfo->cinfo, COL_PROTOCOL, "SoIP");

    if (check_col(pinfo->cinfo, COL_INFO))
        col_clear(pinfo->cinfo, COL_INFO);    /* the COL_INFO is appended every time, so start blank */

    this_msg_len = tvb_length(tvb); /* for safety until we know better */

    while (this_msg_len > offset) {
        offset += dissect_oyster3g_frame(tvb, offset, pinfo, tree, &desegment);
        if (desegment)
            return offset;

        /* If packet contains a follow-on message, add a tab separator */
        if (tvb_length_remaining(tvb,offset)>0) {
            if (check_col(pinfo->cinfo, COL_INFO))
                col_append_str(pinfo->cinfo, COL_INFO, "\t");
        }
    }

    return this_msg_len;
}

static int dissect_oyster3g_frame(tvbuff_t *old_tvb, int old_offset, packet_info *pinfo, proto_tree *main_tree, int *desegment)
{
    guint32 msg_len, this_msg_len;
    guint8 stream;
    int offset=0;
    tvbuff_t *tvb, *next_tvb;

    /* Set up structures needed to add the protocol subtree and manage it */
    proto_item *ti=NULL;
    proto_tree *tree=NULL;

    *desegment = 0;
    this_msg_len = tvb_length_remaining(old_tvb, old_offset); /* for safety until we know better */

    msg_len = tvb_get_ntohs(old_tvb, old_offset);
    if ((msg_len+OYSTER3G_HEADER_LEN > this_msg_len) && (pinfo->can_desegment)) { /* The message is split across segment boundaries */
        /* Tell the TCP dissector where the data for this
         * message starts in the data it handed us, and how many
         * more bytes we need, and return.
         */
        pinfo->desegment_offset = offset;
        pinfo->desegment_len = msg_len + OYSTER3G_HEADER_LEN - this_msg_len;
        *desegment = 1;
        return offset;
    }
    else
        tvb = tvb_new_subset(old_tvb, old_offset, msg_len+OYSTER3G_HEADER_LEN, msg_len+OYSTER3G_HEADER_LEN);

    if (main_tree) {
        /* create display subtree for the protocol */
        ti = proto_tree_add_item(main_tree, proto_oyster3g, tvb, offset, msg_len+OYSTER3G_HEADER_LEN, FALSE);
        tree = proto_item_add_subtree(ti, ett_oyster3g);
    }

    /* start with the 16 bit length field */
    if (tree) {
        if (msg_len < OYSTER3G_HEADER_LEN) {
            ti = proto_tree_add_text(tree, tvb, offset, 2,"Illegal Msg length %d (TCP segmentation ?): Skipping rest", msg_len);
            expert_add_info_format(pinfo, ti, PI_MALFORMED, PI_ERROR, "Illegal Msg length %d (TCP segmentation ?): Skipping rest", msg_len);
		}
        else
            proto_tree_add_uint(tree, hf_msg_len, tvb, offset, 2, msg_len);
    }
    offset += 2;

    /* Next an 8 bit stream Id */
    stream = tvb_get_guint8(tvb, offset);
    if (tree)
        proto_tree_add_uint(tree, hf_stream, tvb, offset, 1, stream);
    offset += 1;

    /* finally a 4 bit version and 4 bit segmentation control value */
    if (tree) { /* the following fields are within the same octet */
        proto_tree_add_item(tree, hf_version, tvb, offset, 1, FALSE);
        proto_tree_add_item(tree, hf_segmentation, tvb, offset, 1, FALSE);
    }
    offset += 1;

    /* construct a subset TVB for the payload to be disected with - assuming it is in this packet */
    next_tvb = tvb_new_subset(tvb, offset, msg_len, msg_len);
    if (tvb_length_remaining(tvb,offset)>0)
        offset += msg_len;

       /* if the stream has been regsitered by another protocol, hook into it, otherwise dissect as data */
    if (!dissector_try_port(subdissector_table, stream, next_tvb, pinfo, main_tree))
        call_dissector(data_handle,next_tvb,pinfo,main_tree);

    if (tree)
        proto_item_set_len(ti,offset);

    return offset;
}

/* Register the protocol */
void
proto_register_oyster3g(void)
{
/* Setup list of header fields */
    static hf_register_info hf[] = {
        { &hf_msg_len,
            { "Message Length", "soip.msg_len",
            FT_UINT16, BASE_DEC, NULL, 0x0,
            "Message Length", HFILL }
        },
        { &hf_stream,
            { "Stream Id", "soip.stream",
            FT_UINT8, BASE_DEC, NULL, 0x0,
            "Stream Identifier", HFILL }
        },
        { &hf_version,
            { "Version", "soip.version",
            FT_UINT8, BASE_DEC, NULL, 0xF0,
            "SoIP Version", HFILL }
        },
        { &hf_segmentation,
            { "Segmentation", "soip.segmentation",
            FT_UINT8, BASE_DEC, VALS(segmentation_vals), 0x0F,
            "Segmentation Control", HFILL }
        }
    };

/* Setup protocol subtree array */
    static gint *ett[] = {
        &ett_oyster3g,
    };

    proto_oyster3g = proto_register_protocol("Oyster3G Signalling over IP", "SoIP", "soip");

    register_dissector("soip", dissect_oyster3g, proto_oyster3g);

/* Required function calls to register the header fields and subtrees used */
    proto_register_field_array(proto_oyster3g, hf, array_length(hf));
    proto_register_subtree_array(ett, array_length(ett));

    /* subdissector code */
    subdissector_table = register_dissector_table("soip.stream", "SoIP Stream", FT_UINT8, BASE_DEC);
}

void
proto_reg_handoff_oyster3g(void)
{
    oyster3g_handle = find_dissector("soip");
    data_handle = find_dissector("data");

#ifdef OYSTER3G_SOIP_TCP_PORT
    dissector_add("tcp.port", OYSTER3G_SOIP_TCP_PORT, oyster3g_handle);
#endif
#ifdef OYSTER3G_SOIP_SSL_PORT
    ssl_dissector_add(OYSTER3G_SOIP_SSL_PORT, "soip", TRUE);
#endif

    heur_dissector_add("tcp", dissect_oyster3g_heur, proto_oyster3g);
}
